"""Includes UserInputService class."""
import abc
import sys
from typing import Union, List, Optional, Dict, TypeVar, Generic

import pandas as pd

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers import list_helper
from organon.fl.core.helpers.data_frame_helper import get_column_native_type
from organon.fl.core.helpers.string_helper import is_null_or_empty
from organon.idq.core.dq_constants import DqConstants
from organon.idq.domain.businessobjects.dq_test_group import DqTestGroup
from organon.idq.domain.businessobjects.record_sources.dq_df_record_source import DqDfRecordSource
from organon.idq.domain.businessobjects.record_sources.dq_file_record_source import DqFileRecordSource
from organon.idq.domain.enums.dq_record_source_type import DqRecordSourceType
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType
from organon.idq.domain.services.base_user_input_validation_service import BaseUserInputValidationService
from organon.idq.domain.services.user_input_validation_service import UserInputValidationService
from organon.idq.domain.settings.abstractions.dq_base_calculation_parameters import DqBaseCalculationParameters
from organon.idq.domain.settings.abstractions.dq_full_process_input import DqFullProcessInput
from organon.idq.domain.settings.calculation.dq_df_calculation_parameters import DqDfCalculationParameters
from organon.idq.domain.settings.calculation.dq_file_calculation_parameters import DqFileCalculationParameters
from organon.idq.domain.settings.comparison.dq_df_comparison_parameters import DqDfComparisonParameters
from organon.idq.domain.settings.comparison.dq_file_comparison_parameters import DqFileComparisonParameters
from organon.idq.domain.settings.date_value_definition import DateValueDefinition
from organon.idq.domain.settings.dq_column_metadata import DqColumnMetadata
from organon.idq.domain.settings.dq_comparison_column_info import DqComparisonColumnInfo
from organon.idq.domain.settings.full_process.dq_df_full_process_input import DqDfFullProcessInput
from organon.idq.domain.settings.full_process.dq_file_full_process_input import DqFileFullProcessInput
from organon.idq.domain.settings.input_source.dq_df_input_source_settings import DqDfInputSourceSettings
from organon.idq.domain.settings.input_source.dq_file_input_source_settings import DqFileInputSourceSettings
from organon.idq.domain.settings.partition_info import PartitionInfo
from organon.idq.services.user_settings.base_dq_user_input import BaseDqUserInput
from organon.idq.services.user_settings.user_date_value_definition import UserDateValueDefinition
from organon.idq.services.user_settings.user_input_source_settings import UserInputSourceSettings
from organon.idq.services.user_settings.user_multi_partition_calculation_params import \
    UserMultiPartitionCalcParams
from organon.idq.services.user_settings.user_partition_info import UserPartitionInfo


def _get_or_default(val, default_val):
    if val is None:
        return default_val
    return val


DqUserInputType = TypeVar("DqUserInputType", bound=BaseDqUserInput)

UserInputValidationServiceType = TypeVar("UserInputValidationServiceType", bound=BaseUserInputValidationService)


class BaseUserInputService(Generic[DqUserInputType], metaclass=abc.ABCMeta):
    """Includes methods to read and validate dq user input"""

    def __init__(self, user_input: DqUserInputType):
        self._user_input = user_input
        self._source_type = self._get_source_type()

    def _validate_user_input(self, source_columns: List[Dict[str, ColumnNativeType]]) -> Dict[str, int]:
        """Validates dq user input and returns columns and their native types in test calculation."""
        service = self._get_input_validation_service(source_columns)
        return service.validate()

    def convert_to_full_process_input(self) -> DqFullProcessInput:
        """Generates DqFullProcessInput from DqUserInput"""
        source_columns = self._get_columns_in_sources()
        self._validate_user_input(source_columns)
        full_proc_inp = self._get_full_proc_input_instance()
        full_proc_inp.calculation_parameters = self._get_calc_params(source_columns)
        full_proc_inp.comparison_parameters = self.__get_comp_params(len(full_proc_inp.calculation_parameters))
        full_proc_inp.use_supplied_calcs_as_comp_inputs = self._user_input.use_supplied_calcs_as_comp_inputs

        return full_proc_inp

    @classmethod
    def _validate_input_source_settings(cls, settings: UserInputSourceSettings):
        UserInputValidationService.validate_calculation_input_source_settings(settings)

    @abc.abstractmethod
    def _get_columns_in_sources(self) -> List[Dict[str, ColumnNativeType]]:
        pass

    def _get_column_native_types(self, settings: UserInputSourceSettings) -> Dict[str, ColumnNativeType]:
        # this method should be called after input source settings validation
        if isinstance(settings.source, str):
            return self.__get_column_native_types_for_file(settings)
        if isinstance(settings.source, pd.DataFrame):
            return self.__get_column_native_types_for_df(settings.source)
        raise ValueError

    @classmethod
    def __get_column_native_types_for_file(cls, settings: UserInputSourceSettings):
        sep = settings.csv_separator if settings.csv_separator is not None else ","
        data_frame = pd.read_csv(settings.source, sep=sep, nrows=100)
        data_frame = data_frame.convert_dtypes()
        return cls.__get_column_native_types_for_df(data_frame)

    @classmethod
    def __get_column_native_types_for_df(cls, data_frame: pd.DataFrame):
        return {col: get_column_native_type(data_frame, col) for col in data_frame.columns}

    @abc.abstractmethod
    def _get_input_validation_service(self, source_columns: List[Dict[str, ColumnNativeType]]) \
            -> UserInputValidationServiceType:
        """Returns service for user input validation"""

    @abc.abstractmethod
    def _get_calc_params(self, source_columns: List[Dict[str, ColumnNativeType]]):
        """Generates and returns list of calculation parameters"""

    def _get_calculation_params_from_user_multi_part_calc_params(self,
                                                                 user_calc_param: UserMultiPartitionCalcParams,
                                                                 all_columns: List[str]) \
            -> List[DqBaseCalculationParameters]:

        partitions = user_calc_param.partitions if user_calc_param.partitions is not None else [None]

        all_calc_params = []
        for i, partition_info in enumerate(partitions):
            calc_param = self._get_calc_param_instance()
            included_columns = self._decide_included_columns(all_columns)
            calc_param.input_source_settings = self._get_input_source_settings(i,
                                                                               user_calc_param.input_source_settings,
                                                                               partition_info,
                                                                               included_columns)
            calc_param.calculation_name = self.__get_calculation_name(
                calc_param.input_source_settings.partition_info_list)

            calc_param.outlier_parameter = _get_or_default(user_calc_param.outlier_parameter,
                                                           DqConstants.OUTLIER_PARAM_DEFAULT)
            calc_param.max_nominal_cardinality_count = \
                _get_or_default(user_calc_param.max_nominal_cardinality_count,
                                DqConstants.MAX_NOMINAL_CARDINALITY_COUNT_DEFAULT)
            calc_param.use_population_nominal_stats = user_calc_param.use_population_nominal_stats
            calc_param.column_dq_metadata_list = self._get_col_metadata_list(
                self._user_input.excluded_column_names,
                self._user_input.column_default_values)
            calc_param.nominal_column_types = DqConstants.NOMINAL_COLUMN_TYPES.copy()
            if self._user_input.include_date_columns is False and \
                    ColumnNativeType.Date in calc_param.nominal_column_types:
                calc_param.nominal_column_types.remove(ColumnNativeType.Date)
            all_calc_params.append(calc_param)

        return all_calc_params

    def _decide_included_columns(self, all_columns: List[str]):
        excluded_columns = self._user_input.excluded_column_names
        if excluded_columns is not None:
            return [col for col in all_columns if col not in excluded_columns]
        return self._user_input.included_column_names

    @staticmethod
    def __get_calculation_name(partition_info_list: List[PartitionInfo]):
        calculation_name = "Calculation"
        if not list_helper.is_null_or_empty(partition_info_list):
            partitions_str = " & ".join([partition_info.to_str() for partition_info in partition_info_list])
            calculation_name = f"Partition-{partitions_str}"
        return calculation_name

    @staticmethod
    def _get_col_metadata_list(excluded_column_names: List[str],
                               column_default_values: Dict[str, List[str]]) \
            -> List[DqColumnMetadata]:
        excluded_column_names = _get_or_default(excluded_column_names, [])
        column_default_values = _get_or_default(column_default_values, {})
        all_columns = set()
        all_columns.update(excluded_column_names)
        all_columns.update(list(column_default_values.keys()))
        ret_list = []
        for col in all_columns:
            inclusion_flag = None
            if excluded_column_names is not None and col in excluded_column_names:
                inclusion_flag = False
            act_meta = DqColumnMetadata()
            act_meta.column_name = col
            act_meta.inclusion_flag = inclusion_flag
            act_meta.default_values = column_default_values[col] if col in column_default_values else None
            ret_list.append(act_meta)
        return ret_list

    @classmethod
    def _get_input_source_settings(cls, index_of_calc: int, settings: UserInputSourceSettings,
                                   partition_info_list: List[UserPartitionInfo],
                                   included_column_names: List[str]) \
            -> Union[DqDfInputSourceSettings, DqFileInputSourceSettings]:

        if isinstance(settings.source, str):
            record_source = DqFileRecordSource(locator=settings.source)
            act_settings = DqFileInputSourceSettings(record_source)
            act_settings.csv_separator = _get_or_default(settings.csv_separator, ",")
            act_settings.number_of_rows_per_step = _get_or_default(settings.number_of_rows_per_step, 1000000)
            act_settings.date_columns = settings.date_columns
        elif isinstance(settings.source, pd.DataFrame):
            name = settings.name
            if is_null_or_empty(name):
                name = f"Input Source of Calculation-{index_of_calc + 1}"
            record_source = DqDfRecordSource(locator=settings.source, name=name)
            act_settings = DqDfInputSourceSettings(record_source)
        else:
            raise ValueError("Invalid input source type!")

        act_settings.sampling_ratio = _get_or_default(settings.sampling_ratio, 1.0)
        act_settings.is_sampling_on = act_settings.sampling_ratio != 1.0
        act_settings.max_num_of_samples = _get_or_default(settings.max_num_of_samples, sys.maxsize)
        act_settings.partition_info_list = cls._get_partition_info_list(partition_info_list)
        act_settings.included_columns = included_column_names
        act_settings.filter_callable = settings.filter_callable
        return act_settings

    @classmethod
    def _get_partition_info_list(cls, user_partition_info_list: List[UserPartitionInfo]) -> List[PartitionInfo]:
        if user_partition_info_list is None:
            return []
        ret_list = []
        for user_info in user_partition_info_list:
            partition_info = PartitionInfo()
            partition_info.column_name = user_info.column_name
            partition_info.column_values = cls.__get_partition_info_column_values(user_info.column_values)
            ret_list.append(partition_info)
        return ret_list

    @classmethod
    def __get_partition_info_column_values(cls, column_values: Union[List[float], List[str],
                                                                     List[UserDateValueDefinition]]):
        new_list = []
        for val in column_values:
            new_val = val
            if isinstance(val, UserDateValueDefinition):
                new_val = DateValueDefinition(year=val.year, month=val.month, day=val.day, hour=val.hour)
            new_list.append(new_val)
        return new_list

    def __get_comp_params(self, calculation_count: int):
        comp_params = self._get_comp_param_instance()
        input_comp = self._user_input
        comp_params.comparison_columns = self.__get_comparison_columns(input_comp.column_benchmark_horizons,
                                                                       input_comp.duplicate_control_columns)
        comp_params.test_groups = BaseUserInputService.__get_test_groups_info(input_comp.excluded_test_groups,
                                                                              input_comp.test_group_benchmark_horizons,
                                                                              calculation_count)
        comp_params.z_score = _get_or_default(input_comp.z_score, DqConstants.Z_SCORE_DEFAULT)
        comp_params.psi_threshold_green = _get_or_default(input_comp.psi_threshold_green,
                                                          DqConstants.PSI_THRESHOLD_GREEN)
        comp_params.psi_threshold_yellow = _get_or_default(input_comp.psi_threshold_yellow,
                                                           DqConstants.PSI_THRESHOLD_YELLOW)
        comp_params.traffic_light_threshold_green = _get_or_default(input_comp.traffic_light_threshold_green,
                                                                    DqConstants.TRAFFIC_LIGHT_THRESHOLD_GREEN)
        comp_params.traffic_light_threshold_yellow = _get_or_default(input_comp.traffic_light_threshold_yellow,
                                                                     DqConstants.TRAFFIC_LIGHT_THRESHOLD_YELLOW)
        comp_params.minimum_cardinality = _get_or_default(input_comp.minimum_cardinality,
                                                          DqConstants.MINIMUM_CARDINALITY_DEFAULT)
        comp_params.maximum_nom_cardinality = _get_or_default(input_comp.maximum_nom_cardinality,
                                                              DqConstants.CONTROL_MAXIMUM_NOM_CARDINALITY_DEFAULT)
        return comp_params

    @staticmethod
    def __get_test_groups_info(excluded_test_groups: List[str],
                               test_group_benchmark_horizons: Dict[str, int],
                               calculation_count: int) -> List[DqTestGroup]:

        default_test_gr_info = BaseUserInputService.__get_default_test_group_info_as_dict(calculation_count)
        if excluded_test_groups is not None:
            for group in excluded_test_groups:
                gr_type = BaseUserInputService.__get_enum_value(group, DqTestGroupType, "test group type")
                default_test_gr_info[gr_type].inclusion_flag = False
        if test_group_benchmark_horizons is not None:
            for group, bmh in test_group_benchmark_horizons.items():
                gr_type = BaseUserInputService.__get_enum_value(group, DqTestGroupType, "test group type")
                default_test_gr_info[gr_type].test_bmh = bmh
        return list(default_test_gr_info.values())

    @staticmethod
    def __get_default_test_group_info_as_dict(calculation_count: int) -> Dict[DqTestGroupType, DqTestGroup]:
        test_groups = [BaseUserInputService.__get_test_group(DqTestGroupType.RULE_BASED_CONTROLS_COLUMN_SET,
                                                             True, 0, calculation_count - 1),
                       BaseUserInputService.__get_test_group(DqTestGroupType.RULE_BASED_CONTROLS_TABLE_SET,
                                                             True, 0, calculation_count - 1),
                       BaseUserInputService.__get_test_group(DqTestGroupType.TABLE_SCHEMA_CONTROLS,
                                                             True, 1, calculation_count - 1),
                       BaseUserInputService.__get_test_group(DqTestGroupType.TRAFFIC_LIGHT_CONTROLS,
                                                             True, 1, calculation_count - 1),
                       BaseUserInputService.__get_test_group(DqTestGroupType.DISTRIBUTIONAL_COMPARISONS,
                                                             True, 1, calculation_count - 1),
                       BaseUserInputService.__get_test_group(DqTestGroupType.MODELLING_AND_CI_CONTROLS_COLUMN_SET,
                                                             True, 5, calculation_count - 1),
                       BaseUserInputService.__get_test_group(DqTestGroupType.MODELLING_AND_CI_CONTROLS_TABLE_SET,
                                                             True, 5, calculation_count - 1)]

        return {info.group_type: info for info in test_groups}

    @staticmethod
    def __get_test_group(gr_type: DqTestGroupType, inclusion_flag: bool, min_bmh: int, test_bmh: int):
        group = DqTestGroup()
        group.group_type = gr_type
        group.inclusion_flag = inclusion_flag
        group.min_bmh = min_bmh
        group.test_bmh = test_bmh
        return group

    @staticmethod
    def __get_enum_value(val_str: str, enum_type, definition: str):
        if val_str is None:
            return None
        valid_strs = [val.name for val in enum_type]
        if val_str not in valid_strs:
            raise ValueError(f"'{val_str}' is not a valid value for {definition}. Valid values are : {valid_strs}")
        return enum_type[val_str]

    @staticmethod
    def __get_comparison_columns(column_benchmark_horizons: Dict[str, int], duplicate_control_columns: List[str]) \
            -> Optional[List[DqComparisonColumnInfo]]:
        comparison_cols = []
        duplicate_control_columns = _get_or_default(duplicate_control_columns, [])
        column_benchmark_horizons = _get_or_default(column_benchmark_horizons, {})
        all_columns = set()
        all_columns.update(list(column_benchmark_horizons.keys()))
        all_columns.update(duplicate_control_columns)
        for col in all_columns:
            act_col = DqComparisonColumnInfo()
            act_col.column_name = col
            act_col.duplicate_column_control = col in duplicate_control_columns
            act_col.benchmark_horizon = column_benchmark_horizons[col] if col in column_benchmark_horizons else None
            comparison_cols.append(act_col)
        return comparison_cols

    @abc.abstractmethod
    def _get_source_type(self) -> DqRecordSourceType:
        """Returns dq source type"""

    def _get_full_proc_input_instance(self):
        source_type = self._source_type
        if source_type == DqRecordSourceType.TEXT:
            inp = DqFileFullProcessInput()
        elif source_type == DqRecordSourceType.DATA_FRAME:
            inp = DqDfFullProcessInput()
        else:
            raise ValueError("Invalid input source type!")
        return inp

    def _get_calc_param_instance(self) -> Union[DqDfCalculationParameters, DqFileCalculationParameters]:
        source_type = self._source_type
        if source_type == DqRecordSourceType.TEXT:
            inp = DqFileCalculationParameters()
        elif source_type == DqRecordSourceType.DATA_FRAME:
            inp = DqDfCalculationParameters()
        else:
            raise ValueError("Invalid input source type!")
        return inp

    def _get_comp_param_instance(self) -> Union[DqFileComparisonParameters, DqDfComparisonParameters]:
        source_type = self._source_type
        if source_type == DqRecordSourceType.TEXT:
            inp = DqFileComparisonParameters()
        elif source_type == DqRecordSourceType.DATA_FRAME:
            inp = DqDfComparisonParameters()
        else:
            raise ValueError("Invalid input source type!")
        return inp



"""Includes BaseUnsupervisedFeatureSelecter class"""
import abc

from organon.ml.feature_selection.services.abstractions.base_feature_selecter import BaseFeatureSelecter


class BaseUnsupervisedFeatureSelecter(BaseFeatureSelecter, metaclass=abc.ABCMeta):
    """Base user service for unsupervised feature selection"""



"""Includes GamClassifier class."""
from typing import List, Union, Dict, Any

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers.data_frame_helper import get_native_type_from_dtype
from organon.fl.mathematics.constants import DOUBLE_MAX
from organon.ml.common.enums.target_type import TargetType
from organon.ml.modelling.algorithms.core.abstractions.base_classifier import BaseClassifier
from organon.ml.modelling.algorithms.core.abstractions.gam_mixin import GamMixin
from organon.ml.preprocessing.helpers import preprocessing_input_helper


class GamClassifier(GamMixin, BaseClassifier):
    """Generalized additive modelling classifier"""

    # pylint: disable=no-member
    def _fit(self, train_data: pd.DataFrame, target_data: Union[pd.DataFrame, pd.Series], **kwargs):
        if isinstance(target_data, pd.DataFrame):
            target_data = target_data[target_data.columns[0]]
        n_unique_classes = target_data.nunique()
        if n_unique_classes > 2:
            raise ValueError("GamClassifier cannot be used for multiclass targets")

        # pylint: disable=attribute-defined-outside-init
        self._target_not_numeric = get_native_type_from_dtype(target_data.dtype) != ColumnNativeType.Numeric
        if self._target_not_numeric:
            if self.target_positive_class is None or self.target_negative_class is None:
                raise ValueError("Target positive and negative classes should be given for non-numeric target data")
            target_data = preprocessing_input_helper.convert_str_series_to_binary(target_data,
                                                                                  self.target_positive_class,
                                                                                  self.target_negative_class)

        self._fit_gam(train_data, target_data)
        return self

    def _predict_proba(self, data: pd.DataFrame) -> pd.DataFrame:
        data = self._get_preprocessed(data)
        classes = list(self._model.classes_)
        if self._target_not_numeric:
            classes = preprocessing_input_helper.convert_binary_series_to_str(pd.Series(classes),
                                                                              self.target_positive_class,
                                                                              self.target_negative_class)
        return pd.DataFrame(self._model.predict_proba(data[self._selected_features]),
                            columns=classes)

    def _predict(self, data: pd.DataFrame) -> pd.DataFrame:
        data = self._get_preprocessed(data)
        prediction = pd.Series(self._model.predict(data[self._selected_features]))
        if self._target_not_numeric:
            prediction = preprocessing_input_helper.convert_binary_series_to_str(prediction, self.target_positive_class,
                                                                                 self.target_negative_class)
        return pd.DataFrame(prediction.values, columns=["prediction"])

    def _get_classes(self) -> List[str]:
        return list(self._model.classes_)

    def _get_target_type(self) -> TargetType:
        return TargetType.BINARY

    def _fit_and_get_final_model(self, train_data: pd.DataFrame, target_data: np.ndarray, keep_list: List[str]):
        regression_model = LogisticRegression()
        regression_model.fit(self._get_cols(train_data, keep_list), target_data)
        return regression_model

    def _reduce_features(self, train_data: pd.DataFrame, target_data: np.ndarray) -> List[str]:
        rejected_list = []
        negative_coef_exists = True
        max_vif = DOUBLE_MAX
        keep_list_all = train_data.columns.tolist()
        while max_vif > self.max_vif_limit or negative_coef_exists:
            keep_list = self._get_keep_list(train_data, target_data, keep_list_all, self._best_alpha)
            vif_data = self._get_vif_data(train_data, keep_list)
            coef_df = self._get_coef_df(train_data, target_data, keep_list)
            negative_coef_exists = sum(coef_df['coef'] < 0) > 0
            max_vif = vif_data["VIF"].max()

            if max_vif > self.max_vif_limit or negative_coef_exists:
                vif_data = pd.merge(vif_data, coef_df, how='left', on='feature')
                to_eliminate = \
                    vif_data[(vif_data['VIF'] > self.max_vif_limit) | (vif_data['coef'] < 0)]["feature"].tolist()

                rejected_list.append(self._get_feature_to_remove(train_data, target_data, keep_list, to_eliminate))
                keep_list_all = [x for x in keep_list_all if x not in rejected_list]
                if len(keep_list_all) == 0:
                    return []
        return keep_list

    @classmethod
    def _get_feature_importances(cls, train_data: pd.DataFrame, target_data: np.ndarray) -> pd.DataFrame:
        rf_classifier = RandomForestClassifier()
        rf_classifier.fit(train_data, target_data)
        return pd.DataFrame(rf_classifier.feature_importances_, columns=['importance'], index=train_data.columns)

    @classmethod
    def _get_coef_df(cls, train_data: pd.DataFrame, target_data: np.ndarray, keep_list: List[str]):
        regression_model = LogisticRegression()
        data = cls._get_cols(train_data, keep_list)
        columns = data.columns.tolist()
        regression_model.fit(data, target_data)
        del data
        coef_df = pd.DataFrame(regression_model.coef_, columns=columns).T
        coef_df = coef_df.reset_index()
        coef_df.columns = ['feature', 'coef']
        return coef_df

    def _get_params_with_defaults(self) -> Dict[str, Any]:
        params = super()._get_params_with_defaults()
        params.update({
            "target_positive_class": None,
            "target_negative_class": None
        })
        return params



"""
This module keeps the helper functions for dictionary type.
"""
from typing import Dict, TypeVar
import logging

from organon.fl.core.helpers import string_helper

T1 = TypeVar('T1')
V1 = TypeVar('V1')


def add_if_key_not_exist(key: T1, value: V1, dict_to_insert: Dict):
    """
    Adds a key-value pair if the key not exists. Raises TypeError otherwise.
    :param key: key
    :param value: value
    :param dict_to_insert: dict to insert the key-value pair
    :return: Nothing
    """
    if key not in dict_to_insert:
        dict_to_insert[key] = value
    else:
        logging.error("Key is already defined in the dictionary!")
        raise TypeError


def camel_case_keys_to_snake_case(_dict: Dict):
    """Converts keys from CamelCase to snake_case """
    new_dict = {}
    for key in _dict:
        val = _dict[key]
        newkey = string_helper.camel_case_to_snake_case(key)
        if isinstance(val, dict):
            new_dict[newkey] = camel_case_keys_to_snake_case(_dict[key])
        else:
            new_dict[newkey] = _dict[key]

    return new_dict



"""todo"""
from typing import List, Optional, Dict, Tuple

from organon.idq.domain.businessobjects.dq_calculation_output import DqCalculationOutput
from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.main_sample_results import MainSampleResults
from organon.idq.domain.businessobjects.record_sources.dq_df_record_source import DqDfRecordSource
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.dq_run_type import DqRunType
from organon.idq.domain.reporting.dq_reporting_executor import DqReportingExecutor
from organon.idq.domain.reporting.objects.base_dq_output_report import BaseDqOutputReport
from organon.idq.domain.services.task_executor.calculation.dq_calculation_task_executor import DqCalculationTaskExecutor
from organon.idq.domain.services.task_executor.comparison.dq_comparison_task_executor import DqComparisonTaskExecutor
from organon.idq.domain.settings.abstractions.dq_base_calculation_parameters import DqBaseCalculationParameters
from organon.idq.domain.settings.abstractions.dq_full_process_input import DqFullProcessInput
from organon.idq.domain.settings.dq_column_metadata import DqColumnMetadata
from organon.idq.domain.settings.dq_comparison_column_info import DqComparisonColumnInfo


class DqFullProcessExecutor:
    """todo"""

    def __init__(self, dq_settings: DqFullProcessInput):
        self.dq_settings = dq_settings
        self._run_type = self._get_run_type()

    def execute(self) -> Tuple[BaseDqOutputReport, List[DqCalculationOutput]]:
        """todo"""
        calculation_results = self._get_calculation_results()
        self.__decide_comparison_columns(calculation_results[-1])
        comparison_output = self._run_controls(calculation_results)
        report = self._generate_reports(comparison_output.comparison_results, calculation_results)
        calculation_outputs = self._get_dq_calculation_outputs(calculation_results)
        return report, calculation_outputs

    def _get_dq_calculation_outputs(self, calculation_results: List[DqCalculationResult]) -> List[DqCalculationOutput]:
        outputs = []
        for i, res in enumerate(calculation_results):
            output = DqCalculationOutput()
            output.calculation_result = res
            res.sample_data = None
            calc_params = self.dq_settings.calculation_parameters[i]
            source = calc_params.input_source_settings.source
            if isinstance(source, DqDfRecordSource):
                source.locator = None
            output.calculation_parameters = self.dq_settings.calculation_parameters[i]
            outputs.append(output)
        return outputs

    def _get_calculation_results(self) -> List[DqCalculationResult]:
        """todo"""
        results = []
        for i, calc in enumerate(self.dq_settings.calculation_parameters):
            is_t_calc = i == len(self.dq_settings.calculation_parameters) - 1  # for t_calc get sample data
            executor = self._get_calculation_task_executor(calc, is_t_calc)
            result = executor.execute()
            result.calculation_name = calc.calculation_name
            results.append(result)
        return results

    def _get_calculation_task_executor(self, calculation_parameters: DqBaseCalculationParameters,
                                       is_t_calculation: bool) \
            -> DqCalculationTaskExecutor:
        return DqCalculationTaskExecutor(calculation_parameters,
                                         run_type=self._run_type, is_test_calculation=is_t_calculation)

    def _run_controls(self, calculation_results: List[DqCalculationResult]) -> MainSampleResults:
        """Runs dq controls on sample data."""
        comparison_executor = self._get_comparison_task_executor(calculation_results)
        comparison_output = comparison_executor.execute()
        return comparison_output

    def _get_comparison_task_executor(self, calculation_results: List[DqCalculationResult]) -> DqComparisonTaskExecutor:
        return DqComparisonTaskExecutor(self.dq_settings.comparison_parameters,
                                        self.dq_settings.calculation_parameters,
                                        calculation_results,
                                        control_types_to_execute=self._get_controls_to_execute())

    def _generate_reports(self, comparison_results: List[DqComparisonResult],
                          calculation_results: List[DqCalculationResult]) -> BaseDqOutputReport:
        """Generates reports from dq comparison results."""
        report_helper = self._get_report_helper(comparison_results, calculation_results)
        report = report_helper.execute()
        return report

    def _get_report_helper(self, comparison_results: List[DqComparisonResult],
                           calculation_results: List[DqCalculationResult]):
        return DqReportingExecutor(self.dq_settings, calculation_results, comparison_results,
                                   self._run_type)

    def _get_controls_to_execute(self) -> Optional[List[DqControlType]]:
        """:returns: controls for run type"""
        common_tests = [DqControlType.EMPTY_TABLE, DqControlType.STABLE_COLUMN, DqControlType.DUPLICATE_KEYS]
        if self._run_type == DqRunType.RUN_ONE_DATA_SOURCE:
            return common_tests
        return common_tests + [DqControlType.UNEXPECTED_NUMERICAL_VALUES, DqControlType.TABLE_COLUMNS,
                               DqControlType.UNEXPECTED_NOMINAL_VALUES, DqControlType.COLUMN_MEAN,
                               DqControlType.DATA_SOURCE_STATS_TL_CONTROL,
                               DqControlType.DATA_SOURCE_STATS_TIME_SERIES_CONTROL, DqControlType.PSI_NOMINAL_VALUES,
                               DqControlType.PSI_NUMERIC_VALUES]

    def _get_run_type(self) -> DqRunType:
        # todo
        if len(self.dq_settings.calculation_parameters) == 1:
            return DqRunType.RUN_ONE_DATA_SOURCE
        return DqRunType.RUN_ONE_DATA_SOURCE_WITH_PARTITIONS

    def __decide_comparison_columns(self, t_calc_result: DqCalculationResult):
        comp_cols_dict: Dict[str, DqComparisonColumnInfo] = \
            {col.column_name: col for col in self.dq_settings.comparison_parameters.comparison_columns}
        meta_dict: Dict[str, DqColumnMetadata] = {col.column_name: col for col in t_calc_result.column_metadata_list}
        duplicate_column_control_default = False
        bh_default = len(self.dq_settings.calculation_parameters) - 1
        for col in t_calc_result.data_source_stats.data_column_collection:
            if col.column_name in meta_dict and meta_dict[col.column_name].inclusion_flag:
                if col.column_name in comp_cols_dict:
                    comp_col = comp_cols_dict[col.column_name]
                    comp_col.duplicate_column_control = duplicate_column_control_default \
                        if comp_col.duplicate_column_control is None else comp_col.duplicate_column_control
                    bh_value = comp_col.benchmark_horizon
                    comp_col.benchmark_horizon = bh_default if bh_value is None or bh_value > bh_default else bh_value
                else:
                    new_comp_col = DqComparisonColumnInfo()
                    new_comp_col.column_name = col.column_name
                    new_comp_col.duplicate_column_control = duplicate_column_control_default
                    new_comp_col.benchmark_horizon = bh_default
                    comp_cols_dict[col.column_name] = new_comp_col
            else:
                if col.column_name in comp_cols_dict:
                    comp_cols_dict.pop(col.column_name)
        self.dq_settings.comparison_parameters.comparison_columns = list(comp_cols_dict.values())



""" This module includes BasePredictor class"""
import abc
from collections import deque
from typing import Tuple, List

import numpy as np

from organon.idq.domain.controls.helpers.algorithms.prediction_result import PredictionResult
from organon.idq.domain.enums.dq_pred_algorithm_type import DqPredAlgorithmType


class BasePredictor(metaclass=abc.ABCMeta):
    """ Base class for prediction"""

    def __init__(self):
        self.predictor_type: DqPredAlgorithmType = None

    @staticmethod
    @abc.abstractmethod
    def predict(input_data: list, t_val: float, n_estimators=200, max_depth=10,
                max_samples_leaf=1) -> PredictionResult:
        """ train and predict data """
        raise NotImplementedError

    @staticmethod
    def generate_train_data_with_lags(past_data: List[float], t_val: float) -> \
            Tuple[np.array, List[float]]:
        """model data generator"""
        input_data = past_data.copy()
        benchmark_horizon = len(input_data)
        input_data.append(t_val)
        labels = input_data
        items = deque(input_data)
        trend_list = deque(range(1, len(input_data) + 1))
        if benchmark_horizon < 5:
            return np.array([trend_list]).T, labels
        if benchmark_horizon < 12:
            lag_num = 1
        elif benchmark_horizon < 24:
            lag_num = 3
        elif benchmark_horizon < 48:
            lag_num = 7
        else:
            lag_num = 14
        lag_list = []
        for _ in range(lag_num):
            items.rotate(1)
            lag_list.append(list(items))
            trend_list.rotate(1)
        return np.append(np.array([list(trend_list)]), np.array(lag_list), 0)[:, lag_num:].T, labels[lag_num:]



"""Includes ObjectClfSettingsValidator class"""
from typing import Tuple

from organon.ml.common.enums.pretrained_model_type import PretrainedModelType
from organon.ml.object_classification.common.helpers import check_directory, throw_val_ex_with_log
from organon.ml.object_classification.common.object_classification_constants import ObjectClassificationConstants
from organon.ml.object_classification.domain.objects.object_clf_settings import ObjectClfSettings


class ObjectClfSettingsValidator:
    """Class for validating ObjectClfSettings"""

    @staticmethod
    def check_settings(settings: ObjectClfSettings):
        """Check directories and image sizes"""
        check_directory(settings.train_data_dir)
        if settings.validation_data_dir is not None:
            check_directory(settings.validation_data_dir)
        ObjectClfSettingsValidator._check_image_size(settings.pretrained_model_settings.model, settings.image_size)

    @staticmethod
    def _check_image_size(model: PretrainedModelType, image_size: Tuple[int, int]):
        size_error = "Using {} model, image_size (width, height) should be bigger than {}"
        err = None
        if model == PretrainedModelType.XCEPTION and (
                image_size[
                    0] <= ObjectClassificationConstants.XCEPTION_MIN_IMG_SIZE or
                image_size[
                    1] <= ObjectClassificationConstants.XCEPTION_MIN_IMG_SIZE):
            err = size_error.format(PretrainedModelType.XCEPTION.name,
                                    ObjectClassificationConstants.XCEPTION_MIN_IMG_SIZE)

        elif model == PretrainedModelType.RES_NET_50_V2 and (
                image_size[
                    0] <= ObjectClassificationConstants.RES_NET_50_V2_MIN_IMG_SIZE or
                image_size[
                    1] <= ObjectClassificationConstants.RES_NET_50_V2_MIN_IMG_SIZE):
            err = size_error.format(PretrainedModelType.RES_NET_50_V2.name,
                                    ObjectClassificationConstants.RES_NET_50_V2_MIN_IMG_SIZE)

        elif model == PretrainedModelType.INCEPTION_RES_NET_V2 and (
                image_size[
                    0] <= ObjectClassificationConstants.INCEPTION_RES_NET_V2_MIN_IMG_SIZE or
                image_size[
                    1] <= ObjectClassificationConstants.INCEPTION_RES_NET_V2_MIN_IMG_SIZE):
            err = size_error.format(PretrainedModelType.INCEPTION_RES_NET_V2.name,
                                    ObjectClassificationConstants.INCEPTION_RES_NET_V2_MIN_IMG_SIZE)
        if err:
            throw_val_ex_with_log(err)



"""todo"""

import abc
import math
from typing import Dict, TypeVar, Generic, List

import pandas as pd

from organon.fl.core.businessobjects.dataframe import DataFrame
from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers import list_helper
from organon.idq.domain.businessobjects.data_column.dq_data_column import DqDataColumn
from organon.idq.domain.businessobjects.dq_data_column_collection import DqDataColumnCollection
from organon.idq.domain.businessobjects.dq_data_source import DqDataSource
from organon.idq.domain.businessobjects.statistics.dq_categorical_statistics import DqCategoricalStatistics
from organon.idq.domain.businessobjects.statistics.sample_statistics import SampleStatistics
from organon.idq.domain.services.statistics.dq_statistics_domain_service import DqStatisticsDomainService
from organon.idq.domain.settings.abstractions.dq_base_calculation_parameters import DqBaseCalculationParameters
from organon.idq.domain.settings.dq_column_metadata import DqColumnMetadata

DqCalcParamsType = TypeVar("DqCalcParamsType", bound=DqBaseCalculationParameters)


class DqSourceStatsBaseService(Generic[DqCalcParamsType], metaclass=abc.ABCMeta):
    """todo"""

    def __init__(self, calc_params: DqCalcParamsType):
        self.calc_params: DqCalcParamsType = calc_params

    @abc.abstractmethod
    def get_data_column_collection(self, data_source: DqDataSource, columns_dq_metadata: List[DqColumnMetadata]) \
            -> DqDataColumnCollection:
        """Generates DqDataColumnCollection instance by acquiring column info from data source and metadata settings"""

    def filter_high_cardinality_columns(self, data_column_collection: DqDataColumnCollection) -> List[str]:
        """todo"""
        col_names = [col.column_name for col in data_column_collection if col.column_native_type
                     in self.calc_params.nominal_column_types]
        included_columns = self.calc_params.input_source_settings.included_columns
        if not list_helper.is_null_or_empty(included_columns):
            col_names = [col for col in col_names if col in included_columns]
        if len(col_names) == 0:
            return []
        cardinality_dict = self._get_column_cardinalities_from_population(col_names)
        return [col for col, cardinality in cardinality_dict.items()
                if cardinality <= self.calc_params.max_nominal_cardinality_count]

    @abc.abstractmethod
    def _get_column_cardinalities_from_population(self, column_names: List[str]) -> Dict[str, int]:
        pass

    def get_population_nominal_statistics(self, nominal_column_names: List[str],
                                          data_col_collection: DqDataColumnCollection) \
            -> Dict[str, DqCategoricalStatistics]:
        """Get statistics for nominal columns in data"""

        filtered_metadata = [col_meta for col_meta in self.calc_params.column_dq_metadata_list
                             if col_meta.inclusion_flag and col_meta.column_name in nominal_column_names]
        dq_columns = {col.column_name: col for col in data_col_collection
                      if col.column_name in [meta.column_name for meta in filtered_metadata]}
        all_stats = {}
        histograms = {}
        if len(dq_columns) > 0:
            histograms = self._get_histograms_for_columns(dq_columns)
        for col_meta in filtered_metadata:
            col_name = col_meta.column_name
            missing_values = col_meta.default_values
            stats = DqCategoricalStatistics(missing_values)
            if col_name in histograms:
                for key, value in histograms[col_name].items():
                    stats.add(key, value)
            all_stats[col_name] = stats
        return all_stats

    @abc.abstractmethod
    def _get_histograms_for_columns(self, dq_data_columns: Dict[str, DqDataColumn]) \
            -> Dict[str, Dict[str, int]]:
        pass

    def get_sample_stats(self, sample_data: pd.DataFrame,
                         data_column_collection: DqDataColumnCollection) -> SampleStatistics:
        """Calculates statistics for calculation input sample"""
        if sample_data is None:
            raise KnownException("Calculation process cancelled because table is empty")
        dataframe = DataFrame()
        dataframe.data_frame = sample_data
        new_col_metadata_list = self.calc_params.column_dq_metadata_list
        filtered_metadata_col = [col for col in new_col_metadata_list if col.inclusion_flag]
        return self.__compute_stats(dataframe, filtered_metadata_col, data_column_collection)

    def __compute_stats(self, data: DataFrame, col_metadata_list: List[DqColumnMetadata],
                        data_column_collection: DqDataColumnCollection) -> SampleStatistics:
        data_columns = {col.column_name: col for col in data_column_collection}
        categorical_cols = []
        numerical_cols = []
        for col in col_metadata_list:
            if col.column_name not in data.get_column_names():
                raise KnownException("Column not in dataframe")
            col_native_type = data_columns[col.column_name].column_native_type
            if col_native_type == ColumnNativeType.Numeric:
                numerical_cols.append(col)
            elif col_native_type in self.calc_params.nominal_column_types:
                categorical_cols.append(col)
        missing_values = {col.column_name: col.default_values for col in categorical_cols}

        stats = DqStatisticsDomainService.compute_categorical_statistics(data, missing_values)
        sample_stats = SampleStatistics()
        sample_stats.nominal_statistics = {col_name: stat for col_name, stat in stats.items()
                                           if stat.cardinality <= self.calc_params.max_nominal_cardinality_count}
        default_values_by_col = {col.column_name: self.__get_default_values(col.default_values)
                                 for col in numerical_cols}
        sample_stats.numerical_statistics = \
            DqStatisticsDomainService.compute_numerical_statistics(data, default_values_by_col,
                                                                   self.calc_params.outlier_parameter)
        return sample_stats

    @staticmethod
    def __get_default_values(default_values: List[str]) -> List[float]:
        float_vals = []
        nan_exists = False
        for val in default_values:
            try:
                float_val = float(val)
            except ValueError:
                continue
            if math.isnan(float_val):
                nan_exists = True
            float_vals.append(float_val)
        if not nan_exists:
            float_vals.append(float("nan"))
        return float_vals

    def update_calc_params(self, col_collection: DqDataColumnCollection) -> List[DqColumnMetadata]:
        """Updates calculation metadata list using column info"""
        return self.__update_column_dq_metadata_list(col_collection)

    def __update_column_dq_metadata_list(self, col_collection: DqDataColumnCollection) -> List[DqColumnMetadata]:
        """Updates calculation metadata list using column info"""
        calc_params = self.calc_params
        col_meta_dict: Dict[str, DqColumnMetadata] = {col.column_name: col for col in
                                                      calc_params.column_dq_metadata_list}
        eligible_types = [ColumnNativeType.Numeric] + calc_params.nominal_column_types
        for col in col_collection:
            if col.column_native_type in eligible_types:
                if col.column_name in col_meta_dict:
                    col_meta = col_meta_dict[col.column_name]
                    col_meta.inclusion_flag = True if col_meta.inclusion_flag is None else col_meta.inclusion_flag
                    col_meta.default_values = [] if list_helper.is_null_or_empty(col_meta.default_values) \
                        else col_meta.default_values
                else:
                    new_meta = DqColumnMetadata()
                    new_meta.column_name = col.column_name
                    new_meta.inclusion_flag = True
                    new_meta.default_values = []
                    col_meta_dict[col.column_name] = new_meta
            else:
                if col.column_name in col_meta_dict:
                    col_meta_dict.pop(col.column_name)
        metadata_list = list(col_meta_dict.values())
        self.calc_params.column_dq_metadata_list = metadata_list
        return metadata_list



"""Includes RFRegressor class."""
from sklearn.ensemble import RandomForestRegressor

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_regressor import BaseSklearnRegressor


class RFRegressor(BaseSklearnRegressor[RandomForestRegressor]):
    """Wrapper for sklearn RandomForestRegressor"""

    @property
    def regressor_class(self):
        return RandomForestRegressor



"""Includes FeatureSimilaritySelection class"""
from typing import Optional

import pandas as pd

from organon.ml.feature_selection.domain.objects.feature_similarity_selection_output import \
    FeatureSimilaritySelectionOutput
from organon.ml.feature_selection.domain.objects.settings.feature_similarity_selection_settings import \
    FeatureSimilaritySelectionSettings
from organon.ml.feature_selection.domain.services.feature_similarity_selection_service import \
    FeatureSimilaritySelectionService
from organon.ml.feature_selection.services.abstractions.base_unsupervised_feature_selecter import \
    BaseUnsupervisedFeatureSelecter


class FeatureSimilaritySelection(BaseUnsupervisedFeatureSelecter):
    """Feature selecter which utilizes a feature similarity algorithm for selection"""

    def __init__(self, data: pd.DataFrame, *, selection_percent: float = 0.05,
                 prioritize_performance: bool = False):
        super().__init__(data)
        self.settings = FeatureSimilaritySelectionSettings(data, selection_percent=selection_percent,
                                                           prioritize_performance=prioritize_performance)
        self.output: Optional[FeatureSimilaritySelectionOutput] = None

    def _run(self):
        service = FeatureSimilaritySelectionService()
        self.output = service.run_selection(self.settings)
        return self.output



"""
This module includes RecordReader class.
"""
import abc

from organon.afe.domain.settings.record_source import RecordSource
from organon.afe.domain.enums.record_source_type import RecordSourceType
from organon.fl.core.exceptionhandling.known_exception import KnownException


class BaseRecordReader(metaclass=abc.ABCMeta):
    """Base class for record reader classes."""

    def __init__(self, record_source: RecordSource):
        self._record_source: RecordSource = record_source

    def read(self):
        """
        Reads the table of records and returns a collection
        :return: A collection containing records in record source
        """
        source_type = self._record_source.get_type()
        if source_type == RecordSourceType.DATABASE:
            return self.read_db_source()
        if source_type == RecordSourceType.TEXT:
            return self.read_text_source()
        if source_type == RecordSourceType.DATA_FRAME:
            return self.read_dataframe_source()
        raise KnownException("Record source type is not valid!")

    @abc.abstractmethod
    def read_text_source(self):
        """Reads text source and returns a collection"""

    @abc.abstractmethod
    def read_dataframe_source(self):
        """Reads DataFrame source and returns a collection"""

    @abc.abstractmethod
    def read_db_source(self):
        "Reads database source and returns a collection"



"""Includes SamplingPreprocessingService class"""
from typing import List

import pandas as pd

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers.data_frame_helper import get_column_native_type
from organon.ml.common.helpers.df_ops_helper import get_bins


class SamplingPreprocessingService:
    """Service to prepare data before sampling"""

    @classmethod
    def discretize(cls, data: pd.DataFrame, columns: List[str], num_bins: int = 10):
        """Converts high cardinality numeric columns to categorical columns with given number of categories(bins)"""
        numeric_columns = [col for col in columns if get_column_native_type(data, col) == ColumnNativeType.Numeric]
        continuous_cols = cls._get_continuous_column_names(data, numeric_columns)
        for col in continuous_cols:
            data[col] = get_bins(data[col], num_bins)
        return data

    @classmethod
    def _get_continuous_column_names(cls, data: pd.DataFrame, columns: List[str]) -> List[str]:
        data_length = len(data)
        upper_limit = data_length * 0.1
        continuous_cols = [col for col in columns if data[col].nunique() > upper_limit]
        return continuous_cols



"""Includes helper functions to use in user input services"""
from typing import Type, TypeVar

from organon.fl.core.exceptionhandling.known_exception import KnownException

T = TypeVar("T")


def get_default_if_none(value: T, default_value: T) -> T:
    """Returns a default value if the value is None, else returns the value"""
    return value if value is not None else default_value


def get_enum(enum_str: str, enum_class: Type[T]) -> T:
    """Converts string to enum """
    if enum_str is None:
        return None

    valid_values = {e.name.upper(): e for e in enum_class}

    if enum_str.upper() in valid_values:
        return valid_values[enum_str.upper()]

    raise KnownException(
        f"Invalid value for enum '{enum_class.__name__}'."
        f"Expected one of: {str(list(valid_values.keys()))} Got: '{enum_str}'")



"""
This module includes AfeColumn class.
"""
from typing import List, Dict, Any

import numpy as np

from organon.afe.domain.enums.afe_operator import AfeOperator
from organon.afe.domain.enums.date_resolution import DateResolution
from organon.afe.domain.settings.afe_date_column import AfeDateColumn
from organon.fl.core.helpers import string_helper


class AfeColumn:
    """
    Keeps metadata for a generated Afe column.
    Columns will be used when building modelling and scoring data.
    """

    def __init__(self):
        self._column_name: str = None
        self.date_column: AfeDateColumn = None
        self.dimension_name: str = None
        self.quantity_name: str = None
        self.operator: AfeOperator = None
        self.group_id: int = None
        self.set: List[np.short] = None
        self.in_out: bool = None
        self.date_resolution: DateResolution = None
        self.offset: int = None
        self.time_window: int = None
        self.source: str = None

    @property
    def column_name(self) -> str:
        """Return value of private attribute '__column_name'"""
        return self._column_name

    def set_column_name(self):
        """
        Sets column name using other attributes
        :return:
        """
        if self.date_column is None:
            date_column_name = ""
        else:
            date_column_name = self.date_column.column_name

        self._column_name = AfeColumn.build_column_name(self.group_id, self.dimension_name, self.quantity_name,
                                                        self.operator.name,
                                                        date_column_name,
                                                        self.time_window)

    def get_group_name(self) -> str:
        """
        Returns group name.
        :return: str
        """
        return f"{self.dimension_name}_{self.quantity_name}_{self.operator.name}_{self.time_window}"

    def get_dimension_set(self, _map: Dict[int, str]) -> str:
        """
        Returns dimension set.
        :type _map: Dict[int, str]
        :return: str
        """
        if len(_map) == 0:
            return ""
        ret_str = ""
        if not self.in_out:
            ret_str += "~"
        for i, elem in enumerate(self.set):
            if _map[elem] is not None:
                token = string_helper.filter_non_alpha_numeric(
                    _map[elem])  # pylint: disable=unsubscriptable-object
                ret_str += token
            else:
                ret_str += "None"

            if i < len(self.set) - 1:
                ret_str += "&"

        return ret_str

    @staticmethod
    def build_column_name(val, dimension_name: str, quantity_name: str, afe_operator_name: str,
                          trx_date_column: str,
                          time_window) -> str:
        """
        Builds and returns column name
        :return: str
        """

        ret_str = f"{dimension_name}_{quantity_name}_{val}_{afe_operator_name}_{trx_date_column}_{time_window}"

        return ret_str

    @staticmethod
    def build_column_name_for_list_value(val: List[Any], dimension_name, quantity_name, afe_op,
                                         time_window) -> str:
        """
        Builds and returns column name
        :return: str
        """

        val_str = "_".join([str(i) for i in val])

        ret_str = f"{dimension_name}_{quantity_name}_{val_str}_{afe_op.name}"

        if time_window is not None:
            ret_str += "_" + str(time_window)

        return ret_str

    def build_column_name_with_map(self, _map: Dict[np.int16, str]) -> str:
        """
        Builds and returns column name using _map
        :param _map:
        :type _map: Dict[int, str]
        :return: column name
        """
        dimension_set = self.get_dimension_set(_map)
        dimension_set_empty = len(dimension_set) == 0

        if dimension_set_empty:
            return f"{self.dimension_name}__{self.quantity_name}__{self.operator.name}__{self.time_window}"
        return f"{self.dimension_name}__{dimension_set}__{self.quantity_name}" \
               f"__{self.operator.name}__{self.time_window}"



"""Includes FeatureReductionService class"""
from typing import List, Dict

import pandas as pd

from organon.ml.feature_reduction.domain.objects.feature_reduction_output import FeatureReductionOutput
from organon.ml.feature_reduction.domain.objects.high_corr_reduction_output import HighCorrReductionOutput
from organon.ml.feature_reduction.domain.objects.similar_dist_reduction_output import SimilarDistReductionOutput
from organon.ml.feature_reduction.domain.objects.univariate_performance_reduction_output import \
    UnivariatePerformanceFeatureReductionOutput
from organon.ml.feature_reduction.domain.reductions.high_correlated_feature_reduction import \
    HighCorrelatedFeatureReduction
from organon.ml.feature_reduction.domain.reductions.null_feature_reduction import NullFeatureReduction
from organon.ml.feature_reduction.domain.reductions.similar_distribution_feature_reduction import \
    SimilarDistributionFeatureReduction
from organon.ml.feature_reduction.domain.reductions.stability_feature_reduction import StabilityFeatureReduction
from organon.ml.feature_reduction.domain.reductions.univariate_performance_feature_reduction import \
    UnivariatePerformanceFeatureReduction
from organon.ml.feature_reduction.settings.objects.base_feature_reduction_settings import BaseFeatureReductionSettings
from organon.ml.feature_reduction.settings.objects.high_correlated_feature_reduction_settings import \
    HighCorrelatedFeatureReductionSettings
from organon.ml.feature_reduction.settings.objects.null_feature_reduction_settings import NullFeatureReductionSettings
from organon.ml.feature_reduction.settings.objects.similar_distribution_feature_reduction_settings import \
    SimilarDistributionFeatureReductionSettings
from organon.ml.feature_reduction.settings.objects.stability_feature_reduction_settings import \
    StabilityFeatureReductionSettings
from organon.ml.feature_reduction.settings.objects.univariate_performance_feature_reduction_settings import \
    UnivariatePerformanceFeatureReductionSettings


class FeatureReductionService:
    """Domain service for feature reduction"""

    @classmethod
    def get_reduction_classes_ordered(cls):
        """Returns all control classes for executor
        """
        # NOTE: Please do not change dict order. Because order is important!
        feature_reduction_dict = {
            NullFeatureReductionSettings: NullFeatureReduction,
            StabilityFeatureReductionSettings: StabilityFeatureReduction,
            UnivariatePerformanceFeatureReductionSettings: UnivariatePerformanceFeatureReduction,
            SimilarDistributionFeatureReductionSettings: SimilarDistributionFeatureReduction,
            HighCorrelatedFeatureReductionSettings: HighCorrelatedFeatureReduction
        }
        return feature_reduction_dict

    @classmethod
    def execute(cls, settings: List[BaseFeatureReductionSettings], data: pd.DataFrame, drop_cols: bool) -> List[
        FeatureReductionOutput]:
        """
        Executes feature reduction with given feature reduction type and settings.
        :param settings:
        :param data:
        :param drop_cols:
        :return:
        """
        feature_reduction_dict = cls.get_reduction_classes_ordered()
        result_list: List[FeatureReductionOutput] = []
        column_univariate_performances: Dict[str, float] = {}
        for feature_reduction_settings, feature_reduction_settings_service in feature_reduction_dict.items():
            for setting in settings:
                if isinstance(setting, feature_reduction_settings):
                    if isinstance(setting, SimilarDistributionFeatureReductionSettings):
                        setting.univariate_performance_result = column_univariate_performances
                    if isinstance(setting, HighCorrelatedFeatureReductionSettings):
                        setting.univariate_performance_result = column_univariate_performances
                    result = feature_reduction_settings_service(setting).execute()
                    res_columns = result.reduced_column_list
                    result_list.append(result)

                    if isinstance(result, UnivariatePerformanceFeatureReductionOutput):
                        column_univariate_performances.update(result.univariate_performance_result)
                    elif isinstance(result, SimilarDistReductionOutput):
                        if result.new_univariate_performance_results is not None:
                            column_univariate_performances.update(result.new_univariate_performance_results)
                    elif isinstance(result, HighCorrReductionOutput):
                        if result.new_univariate_performance_results is not None:
                            column_univariate_performances.update(result.new_univariate_performance_results)

                    if drop_cols and res_columns is not None:
                        data.drop(res_columns, axis=1, inplace=True)

            if len(data.columns) == 0:
                return result_list

        return result_list



"""Includes class DqPopulationNumericalStatistics."""
from typing import List

from organon.fl.core.collections.sorted_dict import SortedDict
from organon.idq.domain.businessobjects.statistics.dq_base_statistics import DqBaseStatistics


class DqPopulationNumericalStatistics(DqBaseStatistics):
    """Data class storing dq population numerical statistics."""

    def __init__(self):
        super().__init__()
        self.missing_values: List[float] = None
        self.missing_values_frequencies: SortedDict[float, int] = None
        self.mean: float = None
        self.std_dev: float = None
        self.variance: float = None
        self.min: float = None
        self.max: float = None
        self.max_cardinality: float = None
        self.max_percentage: float = None
        self.min_cardinality: float = None
        self.min_percentage: float = None



"""Includes helper functions about operating system utilities."""
import os
from typing import Tuple

import psutil

from organon.fl.core.helpers import environment_info_helper
from organon.fl.generic.enums.os_type_id import OsType

OS_TYPE = environment_info_helper.get_os_type()

if OS_TYPE == OsType.LINUX:
    import resource  # pylint: disable=import-error


def get_memory_usage(process: psutil.Process = None) -> int:
    """Returns memory used by given process(current process if argument is None) in bytes"""
    if process is None:
        process = psutil.Process(os.getpid())
    return process.memory_info().rss


def get_peak_memory_usage(process: psutil.Process = None) -> int:
    """Returns peak memory usage of process in bytes"""

    if process is None:
        process = psutil.Process(os.getpid())

    if OS_TYPE == OsType.WINDOWS:
        return process.memory_info().peak_wset
    if OS_TYPE == OsType.LINUX:
        return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
    raise NotImplementedError


def get_cpu_times(process: psutil.Process = None) -> Tuple[float, float]:
    """Returns user/system cpu times utilized by current process"""
    if process is None:
        process = psutil.Process(os.getpid())
    cpu_times = process.cpu_times()
    return cpu_times.user, cpu_times.system


def get_cpu_percentage_since_last_call(process: psutil.Process):
    """Returns cpu utilization percentage of process since last call of this method"""
    return process.cpu_percent() / psutil.cpu_count()



"""Includes DuplicateKeysControl class"""
from typing import List

import pandas as pd

from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.controls.base_dq_control import BaseDqControl
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType
from organon.idq.domain.settings.input_source.dq_df_input_source_settings import DqDfInputSourceSettings
from organon.idq.domain.settings.input_source.dq_file_input_source_settings import DqFileInputSourceSettings


class DuplicateKeysControl(BaseDqControl):
    """Control class for checking if duplicate values exist in primary key columns"""

    @classmethod
    def get_test_group_type(cls) -> DqTestGroupType:
        return DqTestGroupType.RULE_BASED_CONTROLS_COLUMN_SET

    @staticmethod
    def get_control_type() -> DqControlType:
        return DqControlType.DUPLICATE_KEYS

    def get_description(self) -> str:
        return "Duplicate Keys Control"

    def _execute_control(self) -> List[DqComparisonResult]:
        key_columns = [col.column_name for col in self.comp_context.comparison_parameters.comparison_columns
                       if col.duplicate_column_control]
        if len(key_columns) == 0:
            return []

        unique_keys_count = self._get_unique_keys_count(key_columns)
        row_count = self.comp_context.test_calc_result.data_source_stats.row_count
        duplicate_rows = row_count - unique_keys_count
        if duplicate_rows <= 0:
            return []
        source_name = self.comp_context.test_calc_parameters.input_source_settings.source.get_name()
        result = self._get_comparison_result(
            data_entity=DataEntityType.TABLE,
            data_entity_name=source_name,
            property_code="DuplicateKeys",
            result_code=DqComparisonResultCode.TABLE_CONTAINS_DUPLICATE_KEYS,
            message=f"{source_name} contains duplicate keys. Row count: {row_count}. Cardinality: {unique_keys_count}. "
                    f"Number of rows with duplicate keys: {duplicate_rows}"
        )
        result.property_key_numeric = duplicate_rows
        return [result]

    def _get_unique_keys_count(self, key_columns: List[str]):
        input_source_settings = self.comp_context.test_calc_parameters.input_source_settings
        source = input_source_settings.source

        if isinstance(input_source_settings, DqDfInputSourceSettings):
            return len(source.locator.groupby(key_columns).groups)
        if isinstance(input_source_settings, DqFileInputSourceSettings):
            unique_keys = set()
            for dataframe in pd.read_csv(source.locator, sep=input_source_settings.csv_separator,
                                         chunksize=input_source_settings.number_of_rows_per_step):
                unique_keys.update(dataframe.groupby(key_columns).groups.keys())
            return len(unique_keys)
        raise ValueError("InputSourceSettings is not of expected types")



"""
This module includes BinInfo class.
"""
import math

import numpy as np


class BinInfo:
    """
    Information for a histogram Bin
    """

    def __init__(self, *args):

        if len(args) == 0:
            self.count: np.float64 = 0.0
            self.sum: np.float64 = 0.0
            self.sum_of_squares: np.float64 = 0.0
        elif len(args) == 1:
            other: BinInfo = args[0]
            self.initialize(other.count, other.sum, other.sum_of_squares)
        elif len(args) == 3:
            self.initialize(*args)
        else:
            raise NotImplementedError

    def initialize(self, count: np.float64, _sum: np.float64, sum_of_squares: np.float64):
        """Initializes instance with given values"""
        self.count = count
        self.sum = _sum
        self.sum_of_squares = sum_of_squares

    @property
    def average(self) -> np.float64:
        """Returns average. Returns 'nan' if count is not positive"""
        return np.float64(self.count / self.sum) if self.count > 0 else np.float64("nan")

    @property
    def variance(self) -> np.float64:
        """Returns variance. Returns 'nan' if count is not positive"""
        avg = self.average
        if self.count > 0:
            return np.float64(self.sum_of_squares / self.count - avg * avg)
        return np.float64("nan")

    @property
    def std_dev(self) -> np.float64:
        """Returns standard deviation."""
        return np.float64(math.sqrt(self.variance))

    def get_residual_sum_of_squares(self, predicted_average: np.float64 = None) -> np.float64:
        """Returns residual sum of squares.
        :param predicted_average: Predicted average
        :type predicted_average: np.float64
        :return: Residual sum of squares
        """
        avg = self.average
        if predicted_average is not None:
            delta = avg - predicted_average
            if self.count > 0:
                return np.float64(self.sum_of_squares - self.count * avg * avg + self.count * delta * delta)
        else:
            if self.count > 0:
                return np.float64(self.sum_of_squares - self.count * avg * avg)
        return np.float64("nan")

    def overwrite(self, other):
        """Overwrites values 'count,sum,sum_of_squares' with values in given BinInfo
        :param other: Another BinInfo whose values will overwrite this
        :type other: BinInfo
        """
        self.initialize(other.count, other.sum, other.sum_of_squares)

    def __add_subtract(self, count: np.float64, _sum: np.float64, sum_of_squares: np.float64, pos_neg: int = 1):
        """
        Adds values to this
        :param pos_neg: 1 if 'other' will be added -1 if it will be subtracted
        """
        self.count += count * pos_neg
        self.sum += _sum * pos_neg
        self.sum_of_squares += sum_of_squares * pos_neg

    def add(self, *args):
        """
        :param args: Either a single BinInfo or [count(np.float64) and data(np.float64)]
        """
        if len(args) == 1:
            other: BinInfo = args[0]
            self.__add_subtract(other.count, other.sum, other.sum_of_squares)
        elif len(args) == 2:
            count, data = args
            self.__add_subtract(np.float64(count), np.float64(count * data)
                                , np.float64(count * data * data))

    def subtract(self, *args):
        """
        :param args: Either a single BinInfo or [count(np.float64) and data(np.float64)]
        """
        if len(args) == 1:
            other: BinInfo = args[0]
            self.__add_subtract(other.count, other.sum, other.sum_of_squares, pos_neg=-1)
        elif len(args) == 2:
            count, _sum = args
            self.__add_subtract(count, count * _sum, count * _sum * _sum, pos_neg=-1)
        elif len(args) == 3:
            count, _sum, sum_of_squares = args
            self.__add_subtract(count, count * _sum, sum_of_squares, pos_neg=-1)

    def get_metric(self):
        """
        :return: metric
        """
        avg = self.sum / self.count
        return self.count * avg * avg

    @staticmethod
    def distance_between(left, right) -> np.float64:
        """
        :param left:
        :type left: BinInfo
        :param right:
        :type right: BinInfo
        :return: Distance between two BinInfos
        """
        left_mean = left.average
        right_mean = right.average
        left_variance = left.variance
        right_variance = right.variance
        std_dev = math.sqrt(left_variance / left.count + right_variance * right.count)
        return np.float64(math.fabs(left_mean - right_mean) / std_dev)



"""
THis module keeps the OsTypeId enum class.
"""
from enum import Enum


class OsType(Enum):
    """
    This class keeps the operating system types.
    """
    NOT_AVAILABLE = 0
    WINDOWS = 1
    LINUX = 2



"""Includes DictDataFrame class."""
from typing import List, Any, Union, Dict

import numpy as np
import pandas as pd

from organon.fl.core.businessobjects.idataframe import IDataFrame
from organon.fl.core.helpers import string_helper, list_helper


class DictDataFrame(IDataFrame):
    """Dictionary based custom DataFrame implementing IDataFrame abstract class."""

    def __init__(self, row_count: int = None):
        super().__init__(row_count=row_count)
        self.data_frame: Dict[str, np.array] = {}
        self.__columns_marked_for_gc = []

    def set_value(self, value, column_name: str, row_index: int):
        self.data_frame[column_name][row_index] = value

    def contains_column(self, column_name: str) -> bool:
        return column_name in self.data_frame

    def try_add(self, column_name: str, data: List[Any], mark_for_garbage_collection: bool = False) -> bool:
        cond1 = string_helper.is_null_or_empty(column_name) or column_name in self.data_frame
        if cond1:
            raise ValueError("Column name is inadmissible")
        cond2 = list_helper.is_null_or_empty(data) or len(data) != self.row_count
        if cond2:
            raise ValueError("Data length should be equal to frame-row-count")
        if data is not None and len(data) > 0 and isinstance(data[0], str):
            self.data_frame[column_name] = np.array(data)
        else:
            self.data_frame[column_name] = np.array(data, dtype=np.float32)
        if mark_for_garbage_collection:
            self.__columns_marked_for_gc.append(column_name)
        return True

    def remove(self, column_names: Union[List[str], str]):
        columns_to_remove = []
        if isinstance(column_names, list):
            columns_to_remove.extend(column_names)
        else:
            columns_to_remove.append(column_names)
        for col in columns_to_remove:
            del self.data_frame[col]

    def get_value(self, column_name: str, row_index: int = None):
        column = self.data_frame[column_name]
        if row_index is not None:
            return column[row_index]
        return column

    def collect_garbage_columns(self):
        self.remove(self.__columns_marked_for_gc)
        self.__columns_marked_for_gc.clear()

    def get_column_names(self) -> List[str]:
        return list(self.data_frame.keys())

    def get_subset_as_pandas_df(self, indices: List[Any] = None, columns: List[str] = None):
        subset = {}
        if indices is None and columns is None:
            raise ValueError("Both indices and columns are None")

        if indices is None:
            subset = {col: self.data_frame[col] for col in columns}
        elif columns is None:
            for col, values in self.data_frame.items():
                subset[col] = [values[i] for i in indices]
        else:
            for col in columns:
                val = self.data_frame[col]
                subset[col] = [val[i] for i in indices]
        return pd.DataFrame(subset)

    def copy_as_pandas_df(self):
        return pd.DataFrame(self.data_frame)



"""Includes BaseDataSourceHandler class"""
import abc
from typing import TypeVar, Generic, TYPE_CHECKING

import numpy as np
import pandas as pd

from organon.ml.common.enums.classification_type import ClassificationType
from organon.ml.object_classification.common.helpers import throw_val_ex_with_log
from organon.ml.object_classification.domain.objects.data_source_settings import DataSourceSettings

_DataSourceDataTypeT = TypeVar("_DataSourceDataTypeT")

if TYPE_CHECKING:
    import tensorflow as tf  # noqa


class BaseDataSourceHandler(Generic[_DataSourceDataTypeT], metaclass=abc.ABCMeta):
    """Class for handling data different types of data input and outputs"""

    def __init__(self, data: _DataSourceDataTypeT, settings: DataSourceSettings):
        self.data: _DataSourceDataTypeT = data
        self.settings: DataSourceSettings = settings
        self._dataset: "tf.data.Dataset" = None

    @property
    def dataset(self):
        """dataset attribute property method"""
        return self._dataset

    def get_labels(self):
        """Get labels from tensorflow dataset"""
        if self.settings.clf_type == ClassificationType.BINARY:
            return self._get_binary_labels()
        if self.settings.clf_type == ClassificationType.MULTICLASS:
            return self._get_multiclass_labels()
        return throw_val_ex_with_log("Classification type is ambiguous!")

    @abc.abstractmethod
    def process_data(self):
        """Load and transform data"""
        raise NotImplementedError

    @abc.abstractmethod
    def initialize_probs_df(self) -> pd.DataFrame:
        """Initialize probs_df with proper columns"""
        raise NotImplementedError

    @abc.abstractmethod
    def _get_binary_labels(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _get_multiclass_labels(self):
        raise NotImplementedError

    def _extract_labels(self):
        return np.array([y for _, y in self.dataset])

    def _transform_tf_dataset(self):
        import tensorflow as tf # pylint: disable=import-outside-toplevel

        def transform(image, label):
            image = tf.image.convert_image_dtype(image, tf.float32)
            image = tf.image.resize_with_pad(image, target_height=self.settings.image_size[1],
                                             target_width=self.settings.image_size[0])
            image = tf.expand_dims(image, axis=0)
            return image, label

        self._dataset = self._dataset.map(transform)



"""Includes ImputationFitOutput class."""
from typing import Union, Dict

# noinspection PyUnresolvedReferences
from sklearn.experimental import enable_iterative_imputer  # pylint: disable=unused-import
from sklearn.impute import IterativeImputer
from sklearn.impute import SimpleImputer


class ImputationFitOutput:
    """Output of imputation fitting process"""

    def __init__(self):
        self.categorical_imputer: Dict[str, SimpleImputer] = None
        self.numerical_imputer: Union[Dict[str, SimpleImputer], IterativeImputer] = None



"""Includes RegressionReport class."""

import pandas as pd

from organon.ml.reporting.domain.objects.base_report import BaseReport


class RegressionReport(BaseReport):
    """Report for regression"""

    def __init__(self):
        self.performance_summary: pd.DataFrame = None
        self.lift_table: pd.DataFrame = None



"""Includes BinaryReport class."""
import pandas as pd

from organon.ml.reporting.domain.objects.base_report import BaseReport


class BinaryReport(BaseReport):
    """Report for binary classification"""

    def __init__(self):
        self.performance_summary: pd.DataFrame = None
        self.detailed_performance_summary: pd.DataFrame = None
        self.target_based_performance_summary: pd.DataFrame = None



"""Includes Dq user input classes."""
from typing import List, Dict

import pandas as pd

from organon.idq.services.user_settings.user_calculation_params import UserCalculationParams
from organon.idq.services.user_settings.user_date_value_definition import UserDateValueDefinition
from organon.idq.services.user_settings.user_db_obj_locator import UserDbObjLocator
from organon.idq.services.user_settings.user_input_source_settings import UserInputSourceSettings
from organon.idq.services.user_settings.user_multi_partition_calculation_params import \
    UserMultiPartitionCalcParams
from organon.idq.services.user_settings.user_partition_info import UserPartitionInfo


class BaseDqUserInput:
    """User class for all dq settings"""

    # pylint: disable=too-many-instance-attributes
    def __init__(self):
        self.calc_params: UserMultiPartitionCalcParams = None
        self.use_supplied_calcs_as_comp_inputs: bool = None
        self.included_column_names: List[str] = None
        self.excluded_column_names: List[str] = None
        self.column_default_values: Dict[str, List[str]] = None
        self.column_benchmark_horizons: Dict[str, int] = None
        self.duplicate_control_columns: List[str] = None
        self.minimum_cardinality: int = None
        self.maximum_nom_cardinality: int = None
        self.traffic_light_threshold_yellow: float = None
        self.traffic_light_threshold_green: float = None
        self.psi_threshold_yellow: float = None
        self.psi_threshold_green: float = None
        self.z_score: float = None
        self.excluded_test_groups: List[str] = None
        self.test_group_benchmark_horizons: Dict[str, int] = None
        self.include_date_columns: bool = None

    def set_multi_partition_calculation(self,
                                        input_source_settings: dict,
                                        max_nominal_cardinality_count: int = None,
                                        outlier_parameter: float = None,
                                        use_population_nominal_stats: bool = None,
                                        name: str = None,
                                        partitions: List[List[dict]] = None):
        """
        Adds calculation settings
        """
        self._check_simple_types(locals(), {
            "input_source_settings": dict,
            "max_nominal_cardinality_count": int,
            "outlier_parameter": (float, int),
            "name": str,
            "partitions": list
        })  # bu satır method'ta en üst satır olmalı (locals kullanımından dolayı)
        self.calc_params = self._get_multi_partition_calculation_params(name, input_source_settings,
                                                                        max_nominal_cardinality_count,
                                                                        outlier_parameter,
                                                                        use_population_nominal_stats,
                                                                        partitions=partitions)

    def set_comparison_params(self,
                              column_benchmark_horizons: Dict[str, int] = None,
                              duplicate_control_columns: List[str] = None,
                              maximum_nom_cardinality: int = None,
                              minimum_cardinality: int = None,
                              traffic_light_threshold_yellow: float = None,
                              traffic_light_threshold_green: float = None,
                              psi_threshold_yellow: float = None,
                              psi_threshold_green: float = None,
                              z_score: float = None,
                              excluded_test_groups: List[str] = None,
                              test_group_benchmark_horizons: Dict[str, int] = None,
                              included_column_names: List[str] = None,
                              excluded_column_names: List[str] = None,
                              column_default_values: Dict[str, List[str]] = None,
                              include_date_columns: bool = None
                              ):
        """Adds comparison settings"""
        # pylint: disable=too-many-arguments
        self._check_simple_types(locals(),
                                 {
                                     "column_benchmark_horizons": dict,
                                     "duplicate_control_columns": list,
                                     "maximum_nom_cardinality": int,
                                     "minimum_cardinality": int,
                                     "traffic_light_threshold_yellow": [int, float],
                                     "traffic_light_threshold_green": [int, float],
                                     "psi_threshold_yellow": float,
                                     "psi_threshold_green": float,
                                     "z_score": [int, float],
                                     "excluded_test_groups": list,
                                     "test_group_benchmark_horizons": dict,
                                     "included_column_names": list,
                                     "excluded_column_names": list,
                                     "column_default_values": dict,
                                     "include_date_columns": bool,
                                 })

        self.maximum_nom_cardinality = maximum_nom_cardinality
        self.minimum_cardinality = minimum_cardinality
        self.traffic_light_threshold_green = traffic_light_threshold_green
        self.traffic_light_threshold_yellow = traffic_light_threshold_yellow
        self.psi_threshold_green = psi_threshold_green
        self.psi_threshold_yellow = psi_threshold_yellow
        self.z_score = z_score
        self.column_benchmark_horizons = column_benchmark_horizons
        self.duplicate_control_columns = duplicate_control_columns
        self.excluded_test_groups = excluded_test_groups
        self.test_group_benchmark_horizons = test_group_benchmark_horizons
        self.included_column_names = included_column_names
        self.excluded_column_names = excluded_column_names
        self.column_default_values = column_default_values
        self.include_date_columns = include_date_columns

    @classmethod
    def _get_multi_partition_calculation_params(cls, calc_name: str, input_source_settings: dict,
                                                max_nom_card_count: int, outlier_param,
                                                use_pop_nom_stats, partitions: List[List[dict]] = None):
        input_source_settings = cls._get_input_source_settings_from_dict(input_source_settings)

        partitions = cls._get_partitions_from_list(partitions)
        return UserMultiPartitionCalcParams(calc_name, input_source_settings, max_nom_card_count,
                                            outlier_param,
                                            use_pop_nom_stats, partitions)

    @classmethod
    def _get_calculation_params(cls, calc_name: str, input_source_settings, max_nom_card_count,
                                outlier_param,
                                use_pop_nom_stats, partition: List[dict] = None):
        input_source_settings = cls._get_input_source_settings_from_dict(input_source_settings)

        partition = cls._get_partition(partition)
        return UserCalculationParams(calc_name, input_source_settings, max_nom_card_count,
                                     outlier_param,
                                     use_pop_nom_stats, partition=partition)

    @classmethod
    def _get_partitions_from_list(cls, partitions: List[List[dict]]) -> List[List[UserPartitionInfo]]:
        if partitions is not None:
            if not isinstance(partitions, list):
                raise ValueError("partitions should be given as a list of lists")
            all_partition_lists = []
            for i, partition_info_list in enumerate(partitions):
                if not isinstance(partition_info_list, list):
                    raise ValueError("partitions should be given as a list of lists")
                new_list = []
                for j, info in enumerate(partition_info_list):
                    new_info = info
                    if isinstance(info, dict):
                        new_info = BaseDqUserInput._get_partition_info_from_dict(
                            info, f"partitions[{i}][{j}]")
                    new_list.append(new_info)
                all_partition_lists.append(new_list)
            partitions = all_partition_lists
        return partitions

    @classmethod
    def _get_partition(cls, partition_info_list: List[dict]) -> List[UserPartitionInfo]:
        if partition_info_list is None:
            return None
        if not isinstance(partition_info_list, list):
            raise ValueError("A partition should be given as a list of dicts")
        new_list = []
        for j, info in enumerate(partition_info_list):
            new_info = info
            if isinstance(info, dict):
                new_info = BaseDqUserInput._get_partition_info_from_dict(info, f"partition[{j}]")
            new_list.append(new_info)
        return new_list

    @classmethod
    def _get_input_source_settings_from_dict(cls, settings_dict: dict) -> UserInputSourceSettings:
        settings = UserInputSourceSettings()
        BaseDqUserInput._check_simple_types(settings_dict, {
            "source": [UserDbObjLocator, str, pd.DataFrame, dict],
            "name": str,
            "sampling_ratio": float,
            "max_num_of_samples": int,
            "included_columns": list,
            "number_of_rows_per_step": int,
            "partition_info_list": list,
            "date_columns": list
        })

        BaseDqUserInput._get_from_dict(settings_dict, settings, "input_source_settings")

        if settings.filter_callable is not None:
            if not callable(settings.filter_callable) or settings.filter_callable.__code__.co_argcount != 1:
                raise ValueError("'filter_callable' should be a callable which takes "
                                 "only one argument (dataframe to filter)")

        if settings.source is not None:
            if isinstance(settings.source, dict):
                locator = UserDbObjLocator()
                BaseDqUserInput._get_from_dict(settings.source, locator, "input_source_settings['source']")
                settings.source = locator
        return settings

    @staticmethod
    def _get_partition_info_from_dict(info: dict, definition_str: str):
        new_info = UserPartitionInfo()
        BaseDqUserInput._check_simple_types(info, {
            "column_name": str,
            "column_values": list
        }, definition_str)
        base_definition_str = definition_str
        BaseDqUserInput._get_from_dict(info, new_info, base_definition_str)
        if new_info.column_values is not None:
            values_list = []
            for j, value in enumerate(new_info.column_values):
                new_val = value
                if isinstance(value, dict):
                    new_val = UserDateValueDefinition()
                    BaseDqUserInput._check_simple_types(value, {
                        "year": int,
                        "month": int,
                        "day": int,
                        "hour": int
                    }, f"{base_definition_str}['column_values'][{j}]")
                    BaseDqUserInput._get_from_dict(
                        value, new_val, f"{base_definition_str}['column_values'][{j}]")
                values_list.append(new_val)
            new_info.column_values = values_list
        return new_info

    @staticmethod
    def _get_from_dict(val_dict: dict, target_obj: object, definition: str):
        if not isinstance(val_dict, dict):
            raise ValueError(f"{definition} should be given as a dict")
        for key, val in val_dict.items():
            if key not in target_obj.__dict__:
                raise ValueError(f"'{key}' is not a valid attribute for {definition}")
            setattr(target_obj, key, val)

    @staticmethod
    def _get_from_list_of_dicts(val_list: List[dict], target_type: type, elem_definition: str):
        if val_list is None:
            return None
        ret_list = []
        for val_dict in val_list:
            target_obj = target_type()
            BaseDqUserInput._get_from_dict(val_dict, target_obj, elem_definition)
            ret_list.append(target_obj)
        return ret_list

    @staticmethod
    def _check_list_of_dict(_list, description):
        if _list is not None:
            exc = f"{description} should be given as list of dicts."
            if not isinstance(_list, list):
                raise ValueError(exc)
            for val_dict in _list:
                if not isinstance(val_dict, dict):
                    raise ValueError(exc)

    @staticmethod
    def _check_simple_types(val_dict, type_dict, obj_definition: str = None):
        for param, expected_type in type_dict.items():
            param_str = param if obj_definition is None else f"{obj_definition}.{param}"
            if param not in val_dict:
                continue
            val = val_dict[param]
            if val is None:
                continue
            if isinstance(expected_type, list):
                found = False
                for exp_type in expected_type:
                    if isinstance(val, exp_type):
                        found = True
                if not found:
                    type_names = [t.__name__ for t in expected_type]
                    raise ValueError(f"{param_str} should be an instance of one of '{type_names}'")
            else:
                if not isinstance(val, expected_type):
                    raise ValueError(f"{param_str} should be an instance of '{expected_type.__name__}'")



"""Includes DqDataColumnCollection class."""
from typing import List, Optional

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.idq.domain.businessobjects.data_column.dq_data_column import DqDataColumn


class DqDataColumnCollection:
    """Collection of DqDataColumn instances."""

    def __init__(self):
        self.column_list: List[DqDataColumn] = []

    def get_columns(self, col_native_type: ColumnNativeType) -> List[DqDataColumn]:
        """Returns columns with given native type"""
        return [col for col in self.column_list if col.column_native_type == col_native_type]

    def get_column(self, col_name: str) -> Optional[DqDataColumn]:
        """Returns columns with given name. Returns None if not found."""
        for col in self.column_list:
            if col.column_name == col_name:
                return col
        return None

    def get_eligible_columns(self) -> List[DqDataColumn]:
        """Returns eligible columns"""
        return [col for col in self.column_list if col.column_native_type != ColumnNativeType.Other]

    def get_partition_columns(self) -> List[DqDataColumn]:
        """Returns list of columns which are partitionable"""
        return [col for col in self.column_list if col.column_native_type != ColumnNativeType.Other]

    def remove(self, col_name: str):
        """removes column by name"""
        col_to_remove = None
        for col in self.column_list:
            if col.column_name == col_name:
                col_to_remove = col
        if col_to_remove is not None:
            self.column_list.remove(col_to_remove)

    def add(self, item: DqDataColumn):
        """adds item to collection"""
        self.column_list.append(item)

    def __iter__(self):
        return iter(self.column_list)



"""Includes Preprocessor class."""

from typing import Union, List

from organon.fl.mathematics.constants import INT_MAX
from organon.ml.common.enums.target_type import TargetType
from organon.ml.common.helpers.parameter_helper import USE_CLASS_DEFAULT_STR
from organon.ml.preprocessing.domain.services.coarse_class_service import CoarseClassService
from organon.ml.preprocessing.domain.services.imputation_service import ImputationService
from organon.ml.preprocessing.domain.services.one_hot_encoding_service import OneHotEncodingService
from organon.ml.preprocessing.domain.services.scaling_service import ScalingService
from organon.ml.preprocessing.services.user_settings.coarse_input_dto import CoarseInputDto
from organon.ml.preprocessing.settings.enums.imputer_type import ImputerType
from organon.ml.preprocessing.settings.preprocessing_user_input_service import PreprocessingUserInputService


class Preprocessor:
    """User interface class for preprocessing module"""

    @staticmethod
    def get_coarse_class_service(test_ratio: float, target_column_type: str = TargetType.BINARY.name,
                                 min_class_size: int = 1, max_leaf_nodes: int = 20,
                                 stability_check: bool = True,
                                 stability_threshold: float = 0.30, random_state=42,
                                 positive_class: str = None, negative_class: str = None) -> CoarseClassService:
        """returns coarse class service"""
        # pylint: disable=too-many-arguments
        cc_input = CoarseInputDto(test_ratio, target_column_type, min_class_size, max_leaf_nodes, stability_check,
                                  stability_threshold, random_state, positive_class, negative_class)
        cc_settings = PreprocessingUserInputService.get_coarse_class_settings(cc_input)
        return CoarseClassService(cc_settings)

    @staticmethod
    def get_imputation_service(numeric_data_method: str,
                               categorical_data_method: str = ImputerType.SIMPLE.name,
                               n_missing_values: Union[int, float] = USE_CLASS_DEFAULT_STR,
                               c_missing_values: Union[int, float, str] = USE_CLASS_DEFAULT_STR,
                               n_fill_value: Union[float, int] = USE_CLASS_DEFAULT_STR,
                               c_fill_value: Union[str, float, int] = USE_CLASS_DEFAULT_STR,
                               n_strategy: str = USE_CLASS_DEFAULT_STR,
                               c_strategy: str = 'most_frequent',
                               included_columns: List[str] = None) -> ImputationService:
        """
        returns imputation service with given parameters.

        :param data: train data used to fit imputer.
        :param numeric_data_method: type of imputer for numeric columns, "SIMPLE" or "ITERATIVE".
        :param categorical_data_method: type of imputer for categorical columns.
            default: "SIMPLE".
        :param n_missing_values: used to fill numerical imputer missing_values field.
            default: imputer default if None given.
        :param c_missing_values: used to fill categorical imputer missing_values field.
            default: imputer default if None given.
        :param n_fill_value: used to fill numerical imputer fill_value field.
            default: imputer default if None given.
        :param c_fill_value: used to fill categorical imputer fill_value field.
            default: imputer default if None given.
        :param n_strategy: used to fill numerical imputer strategy field.
            default: imputer default if None given.
        :param c_strategy: used to fill categorical imputer strategy field.
            default: "most-frequent"
        :param included_columns: columns to impute
            default: "all columns"
        """
        # pylint: disable=too-many-arguments
        imputation_settings = PreprocessingUserInputService.get_imputation_settings(numeric_data_method,
                                                                                    categorical_data_method,
                                                                                    n_strategy, c_strategy,
                                                                                    n_missing_values, c_missing_values,
                                                                                    n_fill_value, c_fill_value,
                                                                                    included_columns)
        return ImputationService(imputation_settings)

    @staticmethod
    def get_one_hot_encoding_service(category_cumulative_ratio: float = 1.0, category_max_number: int = INT_MAX):
        """
        Returns one hot encoding service with given parameter settings.

        :param float category_cumulative_ratio: cumulative percentage of categories that will be taken as a feature,
        default: 1.0
        :param int category_max_number: maximum number of features to restrict produced features by threshold,
        default: infinite
        """
        one_hot_settings = PreprocessingUserInputService.get_one_hot_encoding_settings(category_cumulative_ratio,
                                                                                       category_max_number)
        service = OneHotEncodingService(one_hot_settings)
        return service

    @staticmethod
    def get_scaling_service(strategy: str):
        """Returns service for scaling numerical columns"""
        scaling_settings = PreprocessingUserInputService.get_scaling_settings(strategy)
        return ScalingService(scaling_settings)



"""This module includes enum class 'FeatureReductionType'"""
from enum import Enum


class FeatureReductionType(Enum):
    """
    Types of Feature Reduction.
    """
    NULL = 1
    STABILITY = 2
    UNIVARIATE_PERFORMANCE = 3
    SIMILAR_DISTRIBUTION = 4
    HIGH_CORRELATION = 5



"""Includes DqDfRecordSource class."""
import pandas as pd

from organon.idq.domain.businessobjects.record_sources.dq_base_record_source import DqBaseRecordSource
from organon.idq.domain.enums.dq_record_source_type import DqRecordSourceType


class DqDfRecordSource(DqBaseRecordSource[pd.DataFrame]):
    """Dq record source for pandas dataframe objects."""

    def __init__(self, locator: pd.DataFrame, name: str):
        super().__init__(locator)
        self.__name = name

    def get_name(self) -> str:
        return self.__name

    @staticmethod
    def get_type() -> DqRecordSourceType:
        return DqRecordSourceType.DATA_FRAME



"""Includes BaseAfeFeature class"""
from typing import Generic, TypeVar

from organon.afe.domain.modelling.businessobjects.afe_column import AfeColumn

AfeColumnType = TypeVar("AfeColumnType", bound=AfeColumn)


class BaseAfeFeature(Generic[AfeColumnType]):
    """Base Afe feature information class"""

    def __init__(self, afe_column: AfeColumnType):
        self.feature_name: str = None
        self.afe_column = afe_column



"""Includes LassoSelectionService class."""
import pandas as pd
from sklearn.linear_model import Lasso

from organon.ml.common.enums.target_type import TargetType
from organon.ml.feature_selection.domain.objects.lasso_selection_output import LassoSelectionOutput
from organon.ml.feature_selection.domain.objects.settings.lasso_selection_settings import LassoSelectionSettings
from organon.ml.feature_selection.domain.services.base_feature_selection_service import SelectionRunSettings
from organon.ml.feature_selection.domain.services.base_supervised_feature_selection_service import \
    BaseSupervisedFeatureSelectionService


class LassoSelectionService(BaseSupervisedFeatureSelectionService[LassoSelectionSettings, LassoSelectionOutput]):
    """Service class for Lasso Selection"""

    @classmethod
    def _validate_settings(cls, settings: LassoSelectionSettings):
        if settings.target_type not in [TargetType.BINARY, TargetType.SCALAR]:
            raise ValueError("Feature selection with Lasso works only for BINARY or SCALAR targets")

    @classmethod
    def _run(cls, run_settings: SelectionRunSettings) -> LassoSelectionOutput:
        settings: LassoSelectionSettings = run_settings.settings
        data = run_settings.data
        target = cls._get_target_data_as_series_of_numbers(settings)

        estimator_args = {} if settings.lasso_args is None else settings.lasso_args
        lasso = Lasso(**estimator_args)
        lasso.fit(data, target)
        lasso_coef_df_tmp = pd.DataFrame(lasso.coef_, columns=['coef'], index=data.columns)
        lasso_coef_df_tmp.reset_index(inplace=True)
        lasso_coef_df_tmp.columns = ['feature', 'coef']
        keep_list = list(lasso_coef_df_tmp[lasso_coef_df_tmp['coef'] != 0]['feature'].values)

        output = LassoSelectionOutput()
        output.selected_features = keep_list
        return output



"""Includes FeatureReductionSettings class."""
from typing import List, Optional
from pandas import DataFrame


class BaseFeatureReductionSettings:
    """Settings for base feature reduction"""
    def __init__(self, data: DataFrame, excluded_columns: Optional[List[str]] = None):
        self.data = data
        self.excluded_columns = excluded_columns



"""Includes SelecterOutput class"""
from typing import List

from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller


class SelecterOutput:
    """Output of Selecter"""

    def __init__(self):
        self.best_modeller: BaseModeller = None
        self.modeller_fold_scores: List[List[float]] = None
        self.modeller_scores: List[float] = None
        self.stacking_fold_scores: List[float] = None
        self.stacking_score: float = None
        self.voting_fold_scores: List[float] = None
        self.voting_score: float = None



"""Includes AlgorithmService class."""
from typing import Union

from organon.fl.core.iocutil import ioc_helper
from organon.ml.common.helpers.user_input_service_helper import get_enum
from organon.ml.modelling.algorithms.core.abstractions.base_classifier import BaseClassifier
from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller
from organon.ml.modelling.algorithms.core.abstractions.base_regressor import BaseRegressor
from organon.ml.modelling.algorithms.core.enums.modeller import Modeller
from organon.ml.modelling.algorithms.core.enums.modeller_type import ModellerType
from organon.ml.modelling.algorithms.services.ml_application_operations import MLApplicationOperations


class AlgorithmService:
    """Service for modelling algorithms"""

    def __init__(self):
        self._initialize_ml()

    @classmethod
    def _initialize_ml(cls):
        MLApplicationOperations.initialize_app()

    @classmethod
    def get_regressor(cls, regressor_type: Union[str, Modeller], **kwargs) -> BaseRegressor:
        """Returns regressor for given regressor type"""
        if isinstance(regressor_type, str):
            regressor_type = get_enum(regressor_type, Modeller)
        if regressor_type.get_modeller_type() != ModellerType.REGRESSOR:
            raise ValueError("Given modeller is not a regression modeller")
        regressor = ioc_helper.resolve(BaseModeller, regressor_type.name, init_args=kwargs)
        return regressor

    @classmethod
    def get_classifier(cls, classifier_type: Union[str, Modeller], **kwargs) -> BaseClassifier:
        """Returns classifier for given classifier type"""
        if isinstance(classifier_type, str):
            classifier_type = get_enum(classifier_type, Modeller)
        if classifier_type.get_modeller_type() != ModellerType.CLASSIFIER:
            raise ValueError("Given modeller is not a classification modeller")
        classifier = ioc_helper.resolve(BaseModeller, classifier_type.name, init_args=kwargs)
        return classifier

    @classmethod
    def get_modeller(cls, modeller_type: Union[str, Modeller], **kwargs) -> BaseModeller:
        """Returns modeller for given modeller type"""
        if isinstance(modeller_type, str):
            modeller_type = get_enum(modeller_type, Modeller)
        modeller = ioc_helper.resolve(BaseModeller, modeller_type.name, init_args=kwargs)
        return modeller



"""Includes UserFeatureReductionSettings class."""


class UserFeatureReductionSettings:
    """Settings for user feature reduction settings"""

    def __init__(self):
        self.data = None
        self.null_ratio_threshold = None
        self.target_type = None
        self.target_column_name = None
        self.performance_metric = None
        self.univariate_performance_threshold = None
        self.correlation_threshold = None
        self.included_reduction_types = None
        self.nunique_count = None
        self.random_state = None
        self.excluded_columns = None



"""Includes DqTaskType"""
from enum import Enum


class DqTaskType(Enum):
    """DQ task types"""
    CALCULATION_TABLE_STATS = 1
    CALCULATION_SAMPLE_STATS = 2
    CALCULATION_NOMINAL_STATS = 3
    CALCULATION_NUMERICAL_STATS = 4
    COMPARISON = 5



"""
This module has functions helping with implementation of afe_settings_reader
"""
import inspect

from typing import List, Dict, Type, get_args, get_origin
from typing import TypeVar

# pylint: disable=invalid-name
T = TypeVar('T')


def get_list_of_objects_from_list_of_dicts(_list: List[Dict], given_class: Type[T]) -> List[T]:
    """
    Given a list of dictionaries storing object attributes and values,this method returns a list of objects of
    type "given_class"

    :param _list: List of dictionaries storing object attributes and their values
    :type _list: List[Dict]
    :param given_class: Type of objects to created and returned in a list
    :return: List[T]
    """
    if isinstance(_list, list):
        ret_list = []
        for obj in _list:
            if not isinstance(obj, given_class) and isinstance(obj, dict):
                ret_list.append(given_class(**obj))
            else:
                ret_list.append(obj)
        return ret_list
    return None


def set_attrs_from_dict(_dict: Dict, obj):
    """
    Set simple type attributes of a given object from a given dictionary storing attributes and their values.
    :param _dict:
    :type _dict: Dict
    :param obj:
    :return:
    """
    for attr in obj.__dict__:
        if attr in _dict:
            _type = type(_dict[attr])
            if _type in [str, int, bool, float, list]:
                setattr(obj, attr, _dict[attr])


def set_defaults_from_dict(obj, defaults_dict: Dict):
    """
    Set attribute values of an object recursively if they are None.
    :param obj: Object to set the attributes
    :param defaults_dict: A dictionary storing the values for attributes
    :type defaults_dict: Dict
    """
    for attr in defaults_dict:
        if attr in obj.__dict__:
            dict_val = defaults_dict[attr]
            obj_val = getattr(obj, attr)
            if obj_val is None:
                if not isinstance(dict_val, dict):
                    setattr(obj, attr, dict_val)
                else:
                    if hasattr(obj, "ATTR_DICT") and obj.ATTR_DICT[attr] == dict:
                        setattr(obj, attr, dict_val)
            elif isinstance(dict_val, dict):
                if isinstance(obj_val, dict):
                    set_defaults_from_dict_to_dict(obj_val, dict_val)
                else:
                    set_defaults_from_dict(obj_val, dict_val)


def set_defaults_from_dict_to_dict(var_dict: Dict, defaults_dict: Dict) -> Dict:
    """
    Assign default values from defaults_dict to var_dict
    :param var_dict: Dictionary to be set by defaults
    :param defaults_dict: Dictionary for default values
    :return: set dictionary
    """

    if var_dict is None:
        return defaults_dict

    for key in defaults_dict:
        var_dict.setdefault(key, defaults_dict[key])

    return var_dict


def setattr_if_none(obj, attr, val):
    """Sets object attribute value to if it is None"""
    if getattr(obj, attr) is None:
        setattr(obj, attr, val)


def get_val_from_kwargs(value, target_type):
    """If values is of target_type, returns value. Else if it is a dict, generates target_type instance from dict. Else
    returns value."""
    if isinstance(value, target_type):
        return value
    if isinstance(value, dict):
        return target_type(**value)
    return value


def get_values_from_kwargs(obj, attr_dict: dict, kwargs_dict: dict, ignored_attrs: List[str] = None):
    """
    Initializes obj using given attributes and values

    :param obj: obj to be initialized
    :param dict attr_dict: map of attribute names and types
    :param dict kwargs_dict: map of attribute values
    :param ignored_attrs: attributes to ignore
    """
    for attr, value in kwargs_dict.items():
        if ignored_attrs is not None and attr in ignored_attrs:
            continue
        if attr not in attr_dict:
            raise ValueError(f"{type(obj)} class has no attribute named '{attr}'")
        attr_type = attr_dict[attr]
        if attr_type != type(value):
            if isinstance(value, dict):
                value = attr_type(**value)
            elif isinstance(value, list):
                if not inspect.isclass(attr_type) and get_origin(attr_type) == list:
                    value = get_list_of_objects_from_list_of_dicts(value, get_args(attr_type)[0])
        setattr(obj, attr, value)



"""todo"""

import abc
from typing import List

from organon.fl.logging.helpers.log_helper import LogHelper
from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode

from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.dq_control_parameters import DqControlParameters
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType


class BaseDqControl(metaclass=abc.ABCMeta):
    """todo"""

    def __init__(self, dq_comparison_context: DqControlParameters):
        self.comp_context = dq_comparison_context

    def execute(self) -> List[DqComparisonResult]:
        """todo"""
        control_description = self.get_description()

        LogHelper.info(f"Executing control: {control_description}")
        results = self._execute_control()

        LogHelper.info(f"Finished executing control: {control_description}")
        return results

    @classmethod
    def _get_comparison_result(cls, data_entity: DataEntityType, data_entity_name: str,
                               result_code: DqComparisonResultCode = None,
                               message: str = None, property_code: str = None):
        return DqComparisonResult(
            data_entity=data_entity, data_entity_name=data_entity_name, test_group=cls.get_test_group_type(),
            result_code=result_code, message=message, property_code=property_code
        )

    @classmethod
    @abc.abstractmethod
    def get_test_group_type(cls) -> DqTestGroupType:
        """return group type of dq control"""

    @staticmethod
    @abc.abstractmethod
    def get_control_type() -> DqControlType:
        """return control type"""

    @abc.abstractmethod
    def _execute_control(self) -> List[DqComparisonResult]:
        pass

    @abc.abstractmethod
    def get_description(self) -> str:
        """todo"""

    def get_test_bmh(self):
        """Returns max benchmark horizon for comparison"""
        return self.comp_context.test_group_info[self.get_test_group_type()].test_bmh



"""
This module includes TargetRecordSource class.
"""
from typing import Union

import pandas as pd

from organon.afe.domain.settings.db_object_input import DbObjectInput
from organon.afe.domain.enums.record_source_type import RecordSourceType
from organon.fl.core.exceptionhandling.known_exception import KnownException


class RecordSource:
    """
    Description of a target record source.
    """

    def __init__(self, **kwargs):
        self.source: Union[DbObjectInput, pd.DataFrame, str] = None

        if "source" in kwargs:
            source = kwargs["source"]
            if isinstance(source, dict):
                self.source = DbObjectInput(**source)
            elif isinstance(source, (str, DbObjectInput, pd.DataFrame)):
                self.source = source
            else:
                raise KnownException("source should be dict, file path string, DbObjectInput or pd.DataFrame")

    def get_type(self) -> RecordSourceType:
        """Returns type of the source"""
        if isinstance(self.source, DbObjectInput):
            return RecordSourceType.DATABASE
        if isinstance(self.source, pd.DataFrame):
            return RecordSourceType.DATA_FRAME
        if isinstance(self.source, str):
            return RecordSourceType.TEXT
        raise KnownException("Record source type is not valid!")



"""
This module includes AfeAlgorithmSettings class.
"""

from organon.afe.domain.common.reader_helper import get_values_from_kwargs


class AfeAlgorithmSettings:
    """
    Base class for algorithm settings
    """
    ATTR_DICT = {
        "dimension_compression_ratio": float,
        "dimension_max_cardinality": int
    }

    def __init__(self, **kwargs):
        self.dimension_compression_ratio: float = None
        self.dimension_max_cardinality: int = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""Includes DqCalculationPersistenceService class."""
from organon.fl.serialization.serialization_helper import serialize_to_file, deserialize_from_file
from organon.idq.core.dq_constants import DqConstants
from organon.idq.domain.businessobjects.dq_calculation_output import DqCalculationOutput


class DqOutputPersistenceService:
    """Class for persisting dq outputs"""

    @classmethod
    def save_calculation_to_file(cls, calculation_output: DqCalculationOutput, location: str):
        """Saves calculation result to given location(file)"""
        serialize_to_file(calculation_output, location, secret_key=DqConstants.get_secret_key())

    @classmethod
    def load_calculation_from_file(cls, location: str, secret_key: str) -> DqCalculationOutput:
        """Loads calculation result from given location(file)"""
        return deserialize_from_file(location, secret_key=secret_key)



"""Includes BaseUnsupervisedFeatureSelectionService class."""
import abc
from typing import TypeVar, Generic

from organon.ml.feature_selection.domain.objects.base_selection_output import BaseSelectionOutput
from organon.ml.feature_selection.domain.objects.settings.base_unsupervised_feature_selection_settings import \
    BaseUnsupervisedFeatureSelectionSettings
from organon.ml.feature_selection.domain.services.base_feature_selection_service import BaseFeatureSelectionService

UnsupervisedFeatureSelectionSettingsType = TypeVar("UnsupervisedFeatureSelectionSettingsType",
                                                   bound=BaseUnsupervisedFeatureSelectionSettings)
FeatureSelectionOutputType = TypeVar("FeatureSelectionOutputType", bound=BaseSelectionOutput)


class BaseUnsupervisedFeatureSelectionService(
    Generic[UnsupervisedFeatureSelectionSettingsType, FeatureSelectionOutputType],
    BaseFeatureSelectionService[UnsupervisedFeatureSelectionSettingsType,
                                FeatureSelectionOutputType],
    metaclass=abc.ABCMeta):
    """Base service class for unsupervised feature selection services"""



"""This module includes TrafficLightComparer class"""
import math
from typing import List

from organon.fl.core.helpers import float_helper
from organon.idq.domain.algorithms.objects.mean_ci_comparison_input import MeanConfidenceIntervalComparisonInput
from organon.idq.domain.algorithms.objects.time_series_comparison_input import TimeSeriesComparisonInput
from organon.idq.domain.algorithms.objects.traffic_light_comparison_input import TrafficLightComparisonInput
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.controls.helpers.algorithms.predictor_factory import PredictorFactory
from organon.idq.domain.controls.helpers.comparison_helper_functions import ComparisonHelperFunctions
from organon.idq.domain.enums.dq_pred_algorithm_type import DqPredAlgorithmType
from organon.idq.domain.enums.signal_type import SignalType


class TrafficLightComparer:
    """Class for TrafficLightComparer"""

    @staticmethod
    def get_dq_comparison_results_for_time_series(comparison_input: TimeSeriesComparisonInput) -> \
            List[DqComparisonResult]:
        """ :returns time series column based comparison results"""
        if TrafficLightComparer.__check_series_all_null(comparison_input.past_series):
            return []
        predictor = PredictorFactory.get_predictor(predictor=DqPredAlgorithmType.RANDOM_FOREST_REG)
        prediction_result = predictor.predict(comparison_input.past_series, comparison_input.current_value)
        nan_indexes = [i for i, val in enumerate(comparison_input.past_series) if math.isnan(val)]
        predictions, actual = prediction_result.predictions, prediction_result.actual
        root_mse = math.sqrt(prediction_result.mean_square_err)

        results = []
        is_current_value_nan = math.isnan(comparison_input.current_value)
        if is_current_value_nan:
            percentage_change = 1
        else:
            percentage_change = ComparisonHelperFunctions.get_percentage_change(predictions[-1], actual[-1])
        tl_signal = ComparisonHelperFunctions.get_traffic_light_signal(
            percentage_change, comparison_input.yellow_threshold, comparison_input.green_threshold)
        add_tl_results = tl_signal is not SignalType.GREEN
        if add_tl_results:
            results.extend(TrafficLightComparer.get_time_series_tl_results(comparison_input, actual, predictions,
                                                                           percentage_change, tl_signal))

        if is_current_value_nan:
            results.extend(TrafficLightComparer.get_time_series_ci_results(comparison_input, actual, predictions,
                                                                           root_mse))
        else:
            add_ci_results = ComparisonHelperFunctions.is_outside_confidence_interval(
                comparison_input.z_score, predictions[-1], root_mse, actual[-1])
            if add_ci_results:
                results.extend(TrafficLightComparer.get_time_series_ci_results(comparison_input, actual, predictions,
                                                                               root_mse))

        if is_current_value_nan:
            results.extend(TrafficLightComparer.get_time_series_ad_results(comparison_input, actual, predictions,
                                                                           float("nan")))
        else:
            abs_errors = TrafficLightComparer.__get_abs_errors(actual, predictions, nan_indexes)
            add_ad_results = abs_errors.index(max(abs_errors)) == len(abs_errors) - 1
            if add_ad_results:
                results.extend(TrafficLightComparer.get_time_series_ad_results(comparison_input, actual, predictions,
                                                                               abs_errors[-1]))

        return results

    @staticmethod
    def __get_abs_errors(actual: List[float], predictions: List[float], nan_indexes: List[int]) -> List[float]:
        abs_errors = []
        for i, val in enumerate(actual):
            predicted = predictions[i]
            if float_helper.is_extreme(val) or float_helper.is_extreme(predicted) or i in nan_indexes:
                err = 0.0
            else:
                err = abs(val - predicted)
            abs_errors.append(err)

        return abs_errors

    @staticmethod
    def __check_series_all_null(series: List[float]):
        for val in series:
            if not math.isnan(val):
                return False
        return True

    @staticmethod
    def get_time_series_ci_results(comparison_input: TimeSeriesComparisonInput, actual: List[float],
                                   predictions: List[float], root_mean_square_err: float) -> \
            List[DqComparisonResult]:
        """ :returns out of confidence interval results"""
        results = []
        result1 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.ci_code,
            property_code="CurrentValue",
            message=comparison_input.ci_message)
        result1.signal_type = SignalType.RED
        result1.property_key_numeric = comparison_input.current_value
        results.append(result1)

        result2 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.ci_code,
            property_code="PredictedValue",
            message=comparison_input.ci_message)
        result2.signal_type = SignalType.RED
        result2.property_key_numeric = predictions[-1]
        results.append(result2)

        result3 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.ci_code,
            property_code="ConfidenceIntervalLowerBound",
            message=comparison_input.ci_message)
        result3.signal_type = SignalType.RED
        result3.property_key_numeric = predictions[-1] - comparison_input.z_score * root_mean_square_err
        results.append(result3)

        result4 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.ci_code,
            property_code="ConfidenceIntervalUpperBound",
            message=comparison_input.ci_message)
        result4.signal_type = SignalType.RED
        result4.property_key_numeric = predictions[-1] + comparison_input.z_score * root_mean_square_err
        results.append(result4)

        for actual_val, pred_val in zip(actual, predictions):
            result = DqComparisonResult(
                data_entity=comparison_input.data_entity,
                data_entity_name=comparison_input.data_entity_name,
                test_group=comparison_input.test_group,
                result_code=comparison_input.ci_code,
                property_code="TimeSeriesAndPredictions",
                message=comparison_input.ci_message)
            result.signal_type = SignalType.RED
            result.property_key_nominal = f"({actual_val}, {pred_val})"
            results.append(result)
        return results

    @staticmethod
    def get_time_series_ad_results(comparison_input: TimeSeriesComparisonInput, actual: List[float],
                                   predictions: List[float], abs_error: float) -> \
            List[DqComparisonResult]:
        """ :returns absolute error maximum results"""
        results = []
        result1 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.ad_code,
            property_code="CurrentValue",
            message=comparison_input.ad_message)
        result1.signal_type = SignalType.RED
        result1.property_key_numeric = comparison_input.current_value
        results.append(result1)

        result2 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.ad_code,
            property_code="PredictedValue",
            message=comparison_input.ad_message)
        result2.signal_type = SignalType.RED
        result2.property_key_numeric = predictions[-1]
        results.append(result2)

        result3 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.ad_code,
            property_code="AbsoluteDeviationBetweenActualAndPredicted",
            message=comparison_input.ad_message)
        result3.signal_type = SignalType.RED
        result3.property_key_numeric = abs_error
        results.append(result3)

        result4 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.ad_code,
            property_code="LastDeviationIsTheMaximumDeviation",
            message=comparison_input.ad_message)
        result4.signal_type = SignalType.RED
        results.append(result4)

        for actual_val, pred_val in zip(actual, predictions):
            result = DqComparisonResult(
                data_entity=comparison_input.data_entity,
                data_entity_name=comparison_input.data_entity_name,
                test_group=comparison_input.test_group,
                result_code=comparison_input.ad_code,
                property_code="TimeSeriesAndPredictions",
                message=comparison_input.ad_message)
            result.signal_type = SignalType.RED
            result.property_key_nominal = f"({actual_val}, {pred_val})"
            results.append(result)
        return results

    @staticmethod
    def get_time_series_tl_results(comparison_input: TimeSeriesComparisonInput, actual: List[float],
                                   predictions: List[float], percentage_change: float,
                                   tl_signal: SignalType) -> List[DqComparisonResult]:
        """ :returns out of traffic light results"""
        results = []
        result1 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.tl_code,
            property_code="CurrentValue",
            message=comparison_input.tl_message)
        result1.signal_type = tl_signal
        result1.property_key_numeric = comparison_input.current_value
        results.append(result1)

        result2 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.tl_code,
            property_code="PredictedValue",
            message=comparison_input.tl_message)
        result2.signal_type = tl_signal
        result2.property_key_numeric = predictions[-1]
        results.append(result2)

        result3 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.tl_code,
            property_code="PercentageDeviationBetweenActualAndPredicted",
            message=comparison_input.tl_message)
        result3.signal_type = tl_signal
        result3.property_key_numeric = percentage_change
        results.append(result3)

        result4 = DqComparisonResult(
            data_entity=comparison_input.data_entity,
            data_entity_name=comparison_input.data_entity_name,
            test_group=comparison_input.test_group,
            result_code=comparison_input.tl_code,
            property_code="SignalType",
            message=comparison_input.tl_message)
        result4.signal_type = tl_signal
        result4.property_key_nominal = tl_signal.name
        results.append(result4)

        for actual_val, pred_val in zip(actual, predictions):
            result = DqComparisonResult(
                data_entity=comparison_input.data_entity,
                data_entity_name=comparison_input.data_entity_name,
                test_group=comparison_input.test_group,
                result_code=comparison_input.tl_code,
                property_code="TimeSeriesAndPredictions",
                message=comparison_input.tl_message)
            result.signal_type = tl_signal
            result.property_key_nominal = f"({actual_val}, {pred_val})"
            results.append(result)
        return results

    @staticmethod
    def get_mi_comparison_results(mi_comparison_input: MeanConfidenceIntervalComparisonInput, control_mean: float,
                                  control_std_err: float) -> List[DqComparisonResult]:
        """ :returns confidence interval results"""
        results = []
        if ComparisonHelperFunctions.is_outside_confidence_interval(mi_comparison_input.z_score,
                                                                    control_mean,
                                                                    control_std_err,
                                                                    mi_comparison_input.current_value):
            result = DqComparisonResult(
                data_entity=mi_comparison_input.data_entity,
                data_entity_name=mi_comparison_input.data_entity_name,
                test_group=mi_comparison_input.test_group,
                result_code=mi_comparison_input.result_code,
                property_code="TestValue",
                message=mi_comparison_input.message)
            result.property_key_numeric = mi_comparison_input.current_value
            results.append(result)

            result = DqComparisonResult(
                data_entity=mi_comparison_input.data_entity,
                data_entity_name=mi_comparison_input.data_entity_name,
                test_group=mi_comparison_input.test_group,
                result_code=mi_comparison_input.result_code,
                property_code="LowerConfidenceLevel",
                message=mi_comparison_input.message)
            result.property_key_numeric = control_mean - mi_comparison_input.z_score * control_std_err
            results.append(result)

            result = DqComparisonResult(
                data_entity=mi_comparison_input.data_entity,
                data_entity_name=mi_comparison_input.data_entity_name,
                test_group=mi_comparison_input.test_group,
                result_code=mi_comparison_input.result_code,
                property_code="UpperConfidenceLevel",
                message=mi_comparison_input.message)
            result.property_key_numeric = control_mean + mi_comparison_input.z_score * control_std_err
            results.append(result)
        return results

    @staticmethod
    def get_tl_comparison_results(tl_comparison_input: TrafficLightComparisonInput, control_mean: float) -> \
            List[DqComparisonResult]:
        """:returns traffic light comparison results"""
        percentage_change = ComparisonHelperFunctions.get_percentage_change(tl_comparison_input.current_value,
                                                                            control_mean)
        signal_type = ComparisonHelperFunctions.get_traffic_light_signal(percentage_change,
                                                                         tl_comparison_input.yellow_threshold,
                                                                         tl_comparison_input.green_threshold)
        results = []
        if signal_type in [SignalType.YELLOW, SignalType.RED]:
            result = DqComparisonResult(
                data_entity=tl_comparison_input.data_entity,
                data_entity_name=tl_comparison_input.data_entity_name,
                test_group=tl_comparison_input.test_group,
                result_code=tl_comparison_input.result_code,
                property_code="TrafficLightSignal",
                message=tl_comparison_input.message)
            result.property_key_nominal = signal_type.name
            results.append(result)

            result = DqComparisonResult(
                data_entity=tl_comparison_input.data_entity,
                data_entity_name=tl_comparison_input.data_entity_name,
                test_group=tl_comparison_input.test_group,
                result_code=tl_comparison_input.result_code,
                property_code="ControlValue",
                message=tl_comparison_input.message)
            result.property_key_numeric = control_mean
            results.append(result)

            result = DqComparisonResult(
                data_entity=tl_comparison_input.data_entity,
                data_entity_name=tl_comparison_input.data_entity_name,
                test_group=tl_comparison_input.test_group,
                result_code=tl_comparison_input.result_code,
                property_code="TestValue",
                message=tl_comparison_input.message)
            result.property_key_numeric = tl_comparison_input.current_value
            results.append(result)

            result = DqComparisonResult(
                data_entity=tl_comparison_input.data_entity,
                data_entity_name=tl_comparison_input.data_entity_name,
                test_group=tl_comparison_input.test_group,
                result_code=tl_comparison_input.result_code,
                property_code="PercentageChange",
                message=tl_comparison_input.message)
            result.property_key_numeric = percentage_change
            results.append(result)
        return results

    @staticmethod
    def get_mi_tl_comparison_results(mi_comparison_input: MeanConfidenceIntervalComparisonInput,
                                     tl_comparison_input: TrafficLightComparisonInput) -> List[DqComparisonResult]:
        """ :returns confidence interval_comparison_result"""
        control_mean, control_std_err = ComparisonHelperFunctions.compute_control_mean_std_avg(
            mi_comparison_input.past_series)
        results = []
        results.extend(
            TrafficLightComparer.get_mi_comparison_results(mi_comparison_input, control_mean, control_std_err))
        results.extend(TrafficLightComparer.get_tl_comparison_results(tl_comparison_input, control_mean))
        return results



"""Includes GamRegressor class."""
from typing import Union, List

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression

from organon.fl.mathematics.constants import DOUBLE_MAX
from organon.ml.common.enums.target_type import TargetType
from organon.ml.modelling.algorithms.core.abstractions.base_regressor import BaseRegressor
from organon.ml.modelling.algorithms.core.abstractions.gam_mixin import GamMixin


class GamRegressor(GamMixin, BaseRegressor):
    """Generalized additive modelling regressor"""

    # pylint:disable=no-member

    def _fit(self, train_data: pd.DataFrame, target_data: Union[pd.DataFrame, pd.Series], **kwargs):
        if isinstance(target_data, pd.DataFrame):
            target_data = target_data[target_data.columns[0]]
        return self._fit_gam(train_data, target_data)

    def _predict(self, data: pd.DataFrame) -> pd.DataFrame:
        data = self._get_preprocessed(data)
        return pd.DataFrame(self._model.predict(data[self._selected_features]), columns=["prediction"])

    def _get_target_type(self) -> TargetType:
        return TargetType.SCALAR

    def _fit_and_get_final_model(self, train_data: pd.DataFrame, target_data: np.ndarray, keep_list: List[str]):
        regression_model = LinearRegression(positive=True)
        regression_model.fit(self._get_cols(train_data, keep_list), target_data)
        return regression_model

    def _reduce_features(self, train_data: pd.DataFrame, target_data: np.ndarray) -> List[str]:
        rejected_list = []
        max_vif = DOUBLE_MAX
        keep_list_all = train_data.columns.tolist()
        while max_vif > self.max_vif_limit:
            keep_list = self._get_keep_list(train_data, target_data, keep_list_all, self._best_alpha)
            vif_data = self._get_vif_data(train_data, keep_list)
            max_vif = vif_data["VIF"].max()
            if max_vif > self.max_vif_limit:
                to_eliminate = vif_data[(vif_data['VIF'] > self.max_vif_limit)]["feature"].tolist()
                rejected_list.append(self._get_feature_to_remove(train_data, target_data, keep_list, to_eliminate))
                keep_list_all = [x for x in keep_list_all if x not in rejected_list]
        return keep_list

    @classmethod
    def _get_feature_importances(cls, train_data: pd.DataFrame, target_data: np.ndarray) -> pd.DataFrame:
        rf_regressor = RandomForestRegressor()
        rf_regressor.fit(train_data, target_data)
        return pd.DataFrame(rf_regressor.feature_importances_, columns=['importance'], index=train_data.columns)



"""Includes BaseSupervisedFeatureSelecter class"""
import abc
from typing import Union

import pandas as pd

from organon.ml.feature_selection.services.abstractions.base_feature_selecter import BaseFeatureSelecter


class BaseSupervisedFeatureSelecter(BaseFeatureSelecter, metaclass=abc.ABCMeta):
    """Base user service for supervised feature selection"""

    def __init__(self, data: pd.DataFrame, target: Union[pd.DataFrame, pd.Series]):
        super().__init__(data)
        self.target: Union[pd.DataFrame, pd.Series] = target

    @classmethod
    def _get_target_as_df(cls, target: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
        target = target if isinstance(target, pd.DataFrame) else pd.DataFrame(target)
        return target



"""Includes TableControlDetails class"""

from organon.idq.core.enums.data_entity_type import DataEntityType


class TableControlDetails:
    """Class for TableControlDetails """

    def __init__(self):
        self.calculation_name: str = None
        self.type: DataEntityType = None
        self.property_key: str = None
        self.value: int = None
        self.state_of_benchmark: int = None

    def to_dict(self):
        """return dict for base alert info"""
        return {
            'calculation_name': self.calculation_name,
            'type': self.type.name,
            'property_key': self.property_key,
            'value': self.value,
            'state_of_benchmark': self.state_of_benchmark,

        }



"""Includes BaseDqOutputReport class."""
import abc
from typing import List

import pandas as pd

from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.dq_data_column_collection import DqDataColumnCollection


class BaseDqOutputReport(metaclass=abc.ABCMeta):
    """Output report for DQ execution."""

    def __init__(self):
        self.alerts_df: pd.DataFrame = None
        self.signals_df: pd.DataFrame = None
        self.signals: List[DqComparisonResult] = None
        self.column_collection: DqDataColumnCollection = None



"""
This module has functions helping with implementation of afe_settings_validator
"""
from typing import List

import numpy as np

from organon.fl.core.exceptionhandling.known_exception import KnownException


def raise_or_add_exception(exc: Exception, exc_list: List[Exception], add_to_list: bool):
    """
    Raises an exception or adds it to list "exc_list" depending on "add_to_list" parameter
    :param exc: Exception to be raised or to be added to exc_list
    :type exc: Exception
    :param exc_list: The list to which exc will be appended if add_to_list is True
    :type exc_list: List[Exception]
    :param add_to_list: If true Exception exc will not be raised and will be added to exc_list
    :type add_to_list: bool
    """
    if add_to_list:
        exc_list.append(exc)
    else:
        raise exc


def check_required_attrs(obj, required_attrs: List[str], obj_str: str = "", exception_list: bool = False) \
        -> List[Exception]:
    """
    Given a list of attribute names of an object, this method checks if these attributes are not None.If there are None
    attributes an exception will be raised or a list of excepitons will be returned.
    :param obj:
    :param required_attrs: Names of attributes to be checked
    :type required_attrs: List[str]
    :param obj_str: A string to define the object instance
    :type obj_str: str
    :param exception_list: If True no exceptions will be raised and a list of excepitons will be returned.
    :type exception_list: bool
    :return:
    """
    exc_list = []
    for attr in required_attrs:
        if getattr(obj, attr) is None:
            exc = KnownException(f"{obj_str}.{attr} is not provided.")
            raise_or_add_exception(exc, exc_list, exception_list)
    return exc_list


def cast_attr(val: str, expected_type: type):
    """
    This function casts a string to types like int,float,bool,numpy.int64...
    :param val: String to be casted
    :type val: str
    :param expected_type: Type to which the string will be casted.
    :type expected_type: type
    :return: Casted value
    """
    if not isinstance(val, str) and expected_type != bool:
        return expected_type(val)

    if expected_type in [int, float, np.int64, np.short, np.float64, np.float32]:
        val = expected_type(val)
    elif expected_type == bool:
        if isinstance(val, bool):
            pass
        elif isinstance(val, str):
            if val.lower() == "true":
                val = True
            elif val.lower() == "false":
                val = False
            else:
                raise ValueError
        else:
            raise TypeError
    else:
        raise TypeError

    return val


def cast_to_np_type(val, np_type):
    """
    Casts a given value to numpy numerical types.
    :param val: Value to be casted
    :param np_type: Numpy type to which the value will be casted
    :return: Casted value.
    """
    if isinstance(val, float):  # prevent 5.5 from being casted to 5
        if np_type not in [np.float32, np.float64]:
            raise TypeError(f"Expected type is float. Got {type(val)}")

    np_types = [np.int64, np.short, np.float32, np.float64]
    if np_type in np_types:
        return np_type(val)

    raise TypeError("np_type should be one of numpy types!")



"""Includes HighCorrReductionOutput class"""
from typing import Dict

from organon.ml.feature_reduction.domain.objects.feature_reduction_output import FeatureReductionOutput


class HighCorrReductionOutput(FeatureReductionOutput):
    """Output of high correlation feature reduction"""

    def __init__(self):
        super().__init__()
        self.new_univariate_performance_results: Dict[str, float] = None



"""
This module keeps the helper functions for string type.
"""
import re

from organon.fl.core.extensions import string_extensions


def filter_non_alpha_numeric(string: str) -> str:
    """
    Returns alpha-numeric characters of the string.
    :param string: input string
    :return: alpha-numeric characters of the input string.
    """
    str_list = []
    for elem in string:
        if str.isalpha(elem) or str.isdigit(elem):
            str_list.append(elem)

    return ''.join(str_list)


def remove_single_character(string: str, char: str) -> str:
    """
    Removes a single character from the string
    :param string: string value
    :param char: character to remove
    :return: new string value without the input char
    """
    str_list = []
    for elem in string:
        if elem != char:
            str_list.append(elem)
    return ''.join(str_list)


def trim_double_quotes(string: str) -> str:
    """
    Trims the double quotes from a string.
    :param string: string value
    :return: trimmed string value
    """
    return remove_single_character(string, "\"")


def equals_ignoring_case(str1: str, str2: str) -> bool:
    """
    Compares if the two strings are equal with ignoring the case.
    :param str1: first string
    :param str2: second string
    :return: bool type to check if the two strings are equal
    """
    if str1 is None and str2 is None:
        return True

    if str1 is None or str2 is None:
        return False
    val = str1.upper() == str2.upper()
    return val


def to_upper_eng(text: str) -> str:
    """
    Returns the upper case representation of the string.
    :param text: text to change
    :return: text with uppercase letters
    """
    val = text.upper()
    return val


def to_lower_eng(text: str) -> str:
    """
    Returns the lower case representation of the string.
    :param text: text to change
    :return: text with lowercase letters
    """
    return text.lower()


def concatenate_all(*args) -> str:
    """
    Concatenates all arguments and returns as an output string.
    :param args: list of objects
    :return: concatenated objects
    """
    output_str: str = ""
    for string in args:
        if string is not None:
            output_str += string

    return output_str


def is_null_or_empty(string: str) -> bool:
    """
    Checks if the string is null or empty.
    :param string: string to be checked
    :return: bool type to check if the string is null or empty
    """
    if (string is None) or (string == ""):
        return True
    return False


def to_null_safe_string(obj: object) -> str:
    """
    Checks if the object is none.
    If so returns none.
    If not returns the string representation of the object.
    :param obj: object to be checked
    :return: string representation of the object
    """
    if obj is None:
        return None
    return str(obj)


def is_null_or_white_space(string: str) -> bool:
    """
    Checks if the string is null or consisting of just white spaces.
    :param string: string to be checked
    :return: bool type to check if the string null or consisting of just white spaces
    """
    return is_null(string) or is_string_just_white_space(string)


def is_null(string: str) -> bool:
    """
    Checks if the string is null.
    :param string: string to be checked
    :return: bool type to check if the string is null.
    """
    return string is None


def is_string_just_white_space(string: str) -> bool:
    """
    Checks if the string consists of just white spaces.
    :param string: string to be checked
    :return: true if the string consists of just white spaces.
    """
    return string.strip() == ""


def concatenate_with_pattern(pattern: str, connector: str, *args) -> str:
    """
    Concatenates all strings with pattern.
    :param pattern: pattern string
    :param connector: connector string
    :param args: list of strings
    :return: string representation of the concatenated strings with a pattern
    """
    if args is None:
        return ""
    concatenated_str: str = ""
    concatenated_str = concatenated_str.join([(pattern + connector).format(arg) for arg in args[:-1]] +
                                             [pattern.format(args[-1])])
    return concatenated_str


def concatenate_objects_with_pattern(pattern: str, connector: str, *args) -> str:
    """
    Concatenates all objects with pattern.
    :param pattern: pattern string
    :param connector: connector string
    :param args: list of objects to concatenate
    :return: string representation of the concatenated objects with a pattern
    """
    if args is None:
        return ""
    concatenated_str: str = ""
    string: str

    for arg in args:
        if arg is None:
            string = "null"
        else:
            string = string_extensions.effective_value(str(arg))
        concatenated_str += pattern.format(string)
        if arg != args[-1]:
            concatenated_str += connector
    return concatenated_str


def concatenate_objects_with_default_pattern(*args) -> str:
    """
    Concatenates all objects with a default pattern and connector.
    :param args: list of objects to concatenate
    :return: string representation of the concatenated objects with the default pattern
    """
    if args is None:
        return ""
    concatenated_str: str = ""
    string: str
    for arg in args:
        if arg is None:
            string = "null"
        else:
            string = string_extensions.effective_value(str(arg))
        concatenated_str += string
        if arg != args[:-1]:
            concatenated_str += ","
    return concatenated_str


camel_case_regex = re.compile(r'([a-z0-9])([A-Z])')


def camel_case_to_snake_case(camel_case_str: str):
    """Converts CamelCase string to snake_case."""
    return camel_case_regex.sub(r'\1_\2', camel_case_str).lower()


def replace_char(string_to_change: str, index: int, new_char: str) -> str:
    """Returns a new string with char at given index replaced by new_char"""
    return f"{string_to_change[:index]}{new_char}{string_to_change[index + 1:]}"



"""Includes Class for Dask Settings"""


class DaskSettings:
    """Class for Dask Settings"""

    def __init__(self):
        self.client = None
        self.use_dask: bool = True
        self.submit_args: dict = None



"""
This module includes IdqApplicationOperations class.
"""
import threading

from organon.fl.generic.interaction.fl_initializer import FlInitializer


class IdqApplicationOperations:
    """Idq Application Operations"""

    APPLICATION_INITIALIZED = False
    __INITIALIZATION_LOCK = threading.Lock()

    @classmethod
    def initialize_app(cls):
        """Initializes application."""
        with IdqApplicationOperations.__INITIALIZATION_LOCK:
            if not IdqApplicationOperations.APPLICATION_INITIALIZED:
                cls._initialize_fl()
                IdqApplicationOperations.register_types()
                IdqApplicationOperations.APPLICATION_INITIALIZED = True

    @classmethod
    def _initialize_fl(cls):
        FlInitializer.application_initialize()

    @classmethod
    def register_types(cls):
        """Registers types to IOC"""
        # todo



"""
This module includes MultiTargetEntityContainer class.
"""
import secrets
from typing import Dict, List

from organon.afe.domain.common import afe_date_helper
from organon.afe.domain.enums.date_resolution import DateResolution
from organon.afe.domain.modelling.businessobjects.target_file_record_collection import TargetFileRecordCollection
from organon.afe.domain.settings.record_source import RecordSource
from organon.fl.core.businessobjects.date_interval import DateInterval
from organon.fl.core.enums.date_interval_type import DateIntervalType
from organon.fl.mathematics import constants


class MultiTargetEntityContainer:
    """
    Container for record collections in every target file record source
    """

    def __init__(self):
        self.__records_per_file: Dict[RecordSource, TargetFileRecordCollection] = {}
        self.__unique_entity_list: List[str] = []

    @property
    def records_per_file(self) -> Dict[RecordSource, TargetFileRecordCollection]:
        """Returns value of private attribute '__records_per_file'"""
        return self.__records_per_file

    @property
    def unique_entity_list(self) -> List[str]:
        """Returns value of private attribute '__unique_entity_list'"""
        return self.__unique_entity_list

    def add(self, source: RecordSource, collection: TargetFileRecordCollection):
        """
        Adds new source-collection pair to records_per_file dictionary
        :param source:
        :type source: RecordSource
        :param collection:
        :type collection: TargetFileRecordCollection
        """
        self.records_per_file[source] = collection

    def remove(self, source: RecordSource):
        """
        Remove source from records_per_file dictionary.
        :param source:
        :type source: RecordSource
        """
        if source in self.records_per_file:
            self.records_per_file.pop(source)

    def unify_entity_list(self, max_sample_size: int):
        """
        Gets unique entity samples from every record source and stores in '__unique_entity_list'
        :param max_sample_size: Maximum number of samples to get from a record source.
        :type max_sample_size: int
        """
        rand = secrets.SystemRandom()
        tmp_list = []
        for collection in self.records_per_file.values():
            tmp_list.extend(collection.get_sampled_unique_entities(rand, max_sample_size))

        self.__unique_entity_list = list(set(tmp_list))

    def get_date_interval_per_entity(self, resolution: DateResolution, max_horizon: int, date_offset: int) -> \
            Dict[str, DateInterval]:
        """
         returns Dictionary where keys are entities and values are DateIntervals
        """
        func_bottom = afe_date_helper.get_date_subtraction_func(resolution, max_horizon + date_offset)
        func_top = afe_date_helper.get_date_subtraction_func(resolution, date_offset)
        tmp: Dict[str, List[int]] = {}
        for collection in self.records_per_file.values():
            for index in range(collection.actual_record_count):
                _id = collection.entities[index]
                event_date = collection.event_dates[index]
                if _id not in tmp:
                    tmp[_id] = [constants.INT_MAX, constants.INT_MIN]
                bottom = func_bottom(event_date)
                top = func_top(event_date)
                if bottom < tmp[_id][0]:
                    tmp[_id][0] = bottom
                if event_date > tmp[_id][1]:
                    tmp[_id][1] = top
        return {key: DateInterval(value[0], value[1], DateIntervalType.CLOSED_CLOSED) for key, value in tmp.items()}



"""Includes NominalColumnControlDetails class"""

from organon.fl.core.enums.column_native_type import ColumnNativeType


class NominalColumnControlDetails:
    """Class for NominalColumnControlDetails """

    def __init__(self):
        self.calculation_name: str = None
        self.column_name: str = None
        self.column_type: ColumnNativeType = None
        self.percentage: float = None
        self.state_of_benchmark: int = None
        self.value: str = None

    def to_dict(self):
        """return dict for base alert info"""
        return {
            'calculation_name': self.calculation_name,
            'column_name': self.column_name,
            'column_type': self.column_type.name,
            'percentage': self.percentage,
            'state_of_benchmark': self.state_of_benchmark,
            'value': self.value,

        }



"""
This module includes AfeReadngSettings class.
"""
from organon.afe.domain.common.reader_helper import get_values_from_kwargs


class AfeDataReadingSettings:
    """Contains settings for record source reading process"""
    ATTR_DICT = {
        "number_of_rows_per_step": int
    }

    def __init__(self, **kwargs):
        self.number_of_rows_per_step: int = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""Includes ObjectClassificationServiceFactory class"""
import os
from typing import Union

from organon.fl.core.fileoperations import directory_helper
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.ml.common.enums.classification_type import ClassificationType
from organon.ml.object_classification.common.helpers import throw_val_ex_with_log
from organon.ml.object_classification.domain.objects.object_clf_settings import \
    ObjectClfSettings
from organon.ml.object_classification.domain.services.binary_object_classificion_service import \
    BinaryObjectClassificationService
from organon.ml.object_classification.domain.services.multiclass_object_classification_service import \
    MulticlassObjectClassificationService


class ObjectClassificationServiceFactory:
    """Class for deciding which service is going to be used in object classification (Binary-Multiclass)"""
    @staticmethod
    def get_classification_service(settings: ObjectClfSettings) -> Union[
            BinaryObjectClassificationService, MulticlassObjectClassificationService]:
        """Returns proper service"""
        train_dir = settings.train_data_dir
        if not directory_helper.exists(train_dir):
            return throw_val_ex_with_log(f"Directory: {train_dir} not found!")

        n_of_classes = len(os.listdir(train_dir))
        if n_of_classes == 2:
            LogHelper.info("Binary classification will start")
            settings.clf_mode = ClassificationType.BINARY
            return BinaryObjectClassificationService(settings)
        if n_of_classes > 2:
            settings.clf_mode = ClassificationType.MULTICLASS
            LogHelper.info("Categorical classification will start")
            return MulticlassObjectClassificationService(settings)
        return throw_val_ex_with_log("At least 2 classes are required for classification!")



"""Includes DqComparisonResultCode enum definition."""
from enum import Enum


class DqComparisonResultCode(Enum):
    """Result code types for dq comparisons"""
    TABLE_IS_EMPTY = 1
    ROW_COUNT_MEAN_IS_OUTSIDE_MEAN_TRAFFIC_LIGHT_INTERVAL = 5
    ROW_COUNT_MEAN_IS_OUTSIDE_MEAN_CONFIDENCE_INTERVAL = 6
    ROW_COUNT_IS_OUTSIDE_PREDICTION_CONFIDENCE_INTERVAL = 7
    ROW_COUNT_IS_OUTSIDE_PREDICTION_TRAFFIC_LIGHT_INTERVAL = 8
    ROW_COUNT_ABSOLUTE_DEVIATION_FROM_PREDICTION_IS_MAXIMUM = 9
    TABLE_SIZE_MEAN_IS_OUTSIDE_MEAN_TRAFFIC_LIGHT_INTERVAL = 10
    TABLE_SIZE_MEAN_IS_OUTSIDE_MEAN_CONFIDENCE_INTERVAL = 11
    TABLE_SIZE_IS_OUTSIDE_PREDICTION_CONFIDENCE_INTERVAL = 12
    TABLE_SIZE_IS_OUTSIDE_PREDICTION_TRAFFIC_LIGHT_INTERVAL = 13
    TABLE_SIZE_ABSOLUTE_DEVIATION_IS_MAXIMUM = 14

    # column level
    COLUMN_ADDED = 2
    COLUMN_REMOVED = 3
    COLUMN_TYPE_CHANGED = 4
    NUMERIC_VALUES_ARE_OUTSIDE_THE_RANGE = 15
    NOMINAL_VALUES_ARE_OUTSIDE_ALLOWED_SET = 16
    FORBIDDEN_NOMINAL_VALUE_IN_DATA = 17
    CARDINALITY_EXCEEDS_MAXIMUM_CARDINALITY = 18
    COLUMN_VALUES_ARE_ALL_ZERO = 19
    COLUMN_VALUES_ARE_ALL_NULL = 20
    UNEXPECTED_NOMINAL_VALUES = 21
    UNEXPECTED_MINIMUM = 22
    UNEXPECTED_MAXIMUM = 23
    # Column-level: Psi, average predictions
    PSI_YELLOW_SIGNAL = 24
    PSI_RED_SIGNAL = 25
    COLUMN_MEAN_IS_OUTSIDE_MEAN_TRAFFIC_LIGHT_INTERVAL = 26
    COLUMN_MEAN_IS_OUTSIDE_MEAN_CONFIDENCE_INTERVAL = 27
    COLUMN_MEAN_IS_OUTSIDE_PREDICTION_CONFIDENCE_INTERVAL = 28
    COLUMN_MEAN_IS_OUTSIDE_PREDICTION_TRAFFIC_LIGHT_INTERVAL = 29
    COLUMN_MEAN_ABSOLUTE_DEVIATION_FROM_PREDICTION_IS_MAXIMUM = 30
    CARDINALITY_IS_BELOW_MINIMUM_THRESHOLD = 31
    COLUMN_IS_CONSTANT = 32
    COLUMNN_VALUES_CONSIST_OF_DEFAULT_VALUES_ONLY = 33
    TABLE_CONTAINS_DUPLICATE_KEYS = 34
    STRATA_ADDED = 35
    STRATA_REMOVED = 36
    COLUMN005_IS_OUTSIDE_MEAN_TRAFFIC_LIGHT_INTERVAL = 37
    COLUMN005_IS_OUTSIDE_MEAN_CONFIDENCE_INTERVAL = 38
    COLUMN005_IS_OUTSIDE_PREDICTION_CONFIDENCE_INTERVAL = 39
    COLUMN005_IS_OUTSIDE_PREDICTION_TRAFFIC_LIGHT_INTERVAL = 40
    COLUMN005_ABSOLUTE_DEVIATION_FROM_PREDICTION_IS_MAXIMUM = 41
    COLUMN995_IS_OUTSIDE_MEAN_TRAFFIC_LIGHT_INTERVAL = 42
    COLUMN995_IS_OUTSIDE_MEAN_CONFIDENCE_INTERVAL = 43
    COLUMN995_IS_OUTSIDE_PREDICTION_CONFIDENCE_INTERVAL = 44
    COLUMN995_IS_OUTSIDE_PREDICTION_TRAFFIC_LIGHT_INTERVAL = 45
    COLUMN995_ABSOLUTE_DEVIATION_FROM_PREDICTION_IS_MAXIMUM = 46



"""Includes UserSamplingSettings class"""
from dataclasses import dataclass
from typing import List

import pandas as pd


@dataclass
class UserSamplingSettings:
    """Stores sampling settings entered by user"""
    data: pd.DataFrame
    target_column_name: str
    target_type: str
    strata_columns: List[str]
    split_test_data: bool
    test_split_ratio: float
    sampling_strategy: str
    sampling_ratio: float
    data_sample_ratio: float



"""Includes StabilityFeatureReduction class."""
from organon.ml.feature_reduction.settings.objects.base_feature_reduction_settings import BaseFeatureReductionSettings

class StabilityFeatureReductionSettings(BaseFeatureReductionSettings):
    """Settings for stability feature reduction"""



"""Includes ImputationSettings class."""
from dataclasses import dataclass
from typing import Union, List

from organon.ml.preprocessing.settings.enums.imputer_type import ImputerType


@dataclass
class ImputationSettings:
    """Settings for Imputation service"""
    numeric_data_method: ImputerType
    categorical_data_method: ImputerType
    n_missing_values: Union[int, float, None]
    c_missing_values: Union[int, float, str, None]
    n_strategy: str
    c_strategy: str
    n_fill_value: Union[float, int]
    c_fill_value: Union[str, float, int]
    included_columns: List[str] = None
